
# STARTING UP:
#
# The suggested way of running this code is the following sequence of commands
#
# > source('loadmain.R')
# > options(error=recover)
# > main(task)
#
# where 'task' is:
# 1 - basic test case
# 2 - curve of median runtime as a function of number of variables
# 3 - curve of presences/absences recovered as a function of maximum conditioning set size
# 4 - presences and absences discovered as a function of assumptions made
#
# Note that the main() function is defined in 'loadmain.R' and sources all .R files in this
# directory before calling the actual runmain() function defined below. This means that any
# saved edits to .R files (including main.R) are automatically sourced, and no further 
# source commands need to be issued after edits. (This gives somewhat 'matlab-like' editing 
# behavior in R.)
#
# See the accompanying README.txt file (in this directory) for details.
#

# this gets called by main() from prompt (after sourcing the loadmain.R file)
runmain <- function( task, theseed=-1 ) {

  cat('--------------------------------------------------------\n')

  # Randomly/manually set seed for reproducibility
  # (omit 'theseed' for random seed; specify a 'theseed' value different from -1 to 
  # start from a specific seed of your choice)
  
  if (theseed != -1) {
    set.seed( theseed )
    cat('Using seed: ',theseed,'\n')
  }
  else {
    theseed <- floor(runif(1,0,1)*100000)
    cat('Using seed: ',theseed,'\n')
    set.seed( theseed )
  }

  #------------------------------------------------------------------------
  # Task 1: Generate a random model and do the inference
  #------------------------------------------------------------------------

  if (task==1) {

    # PARAMETERS

    # True model parameters
    n <- 6
    nexp <- 8
    pdir <- 0.2
    pconf <- 0.2
    acyclic <- 0
    noconfounders <- 0
    
    # Limit the maximum size of the intervention set? (-1 denotes no limit, i.e.
    # it sets the limit to n-2)
    maxcset <- -1
    
    # Assume acyclic or assume no confounders when doing the inference?
    assumeacyclic <- 0
    assumenoconfounders <- 0
    
    # GENERATE THE MODEL AND WRITE IT TO FILE
    
    cat('Generating the model.\n')
    
    # First the direct effects matrix
    Btrue <- matrix( (runif(n*n)<pdir)*1, n, n )
    diag(Btrue) <- 0
    
    # Next the confounder matrix
    Ctrue <- matrix( (runif(n*n)<pconf)*1, n, n )
    Ctrue[lower.tri(Ctrue)] <- 0
    diag(Ctrue) <- 0
    Ctrue <- Ctrue + t(Ctrue)
    
    # Then the experimental setup
    E <- matrix( floor(runif(n*nexp)*3), n, nexp )
    
    # If it was requested that the true model is acyclic, then make it 
    # such (also permute variables to make sure that 1...n isn't a guaranteed
    # valid causal order)
    if (acyclic)
    {
      Btrue[upper.tri(Btrue)] <- 0
      varorder <- sample.int( n, n )
      Btrue <- Btrue[varorder, varorder]
    }
    
    # If it was requrested that the true model has no confounders, then do this.
    if (noconfounders)
    {
      Ctrue <- matrix( 0, n, n )
    }
    
    # Computing the total effects
    Btottrue <- matrix( 0, n, n )
    diag(Btottrue) <- 1
    for (i in 1:n) Btottrue <- (Btottrue | (Btrue %*% Btottrue))*1
    
    # Write the model to a file
    cat('Writing the model file.\n')
    fname <- 'truemodelfile.txt'
    writemodelfile( fname, Btrue, Ctrue, E )

    # CALL DISCOVERER TO DO THE INFERENCE
    
    # Call the C program and get the output
    cat('Calling the C program.\n')
    out <- system(paste('./discoverer3',fname,maxcset,
                  assumeacyclic, assumenoconfounders,'>out.txt',sep=' '))        
    if (out)
    {
      # Some problem occurred. Inform the user
      stop('The discoverer3 program gave a non-zero return value! Something went wrong!')
    }

    cat('Done! Resuming operation in R.\n')    
    
    # SHOW AND ANALYZE THE RESULT
        
    # Source the file
    source('out.txt', local = TRUE)

    # Look at the result
    cat('Emat = \n')
    print(E)
    cat('Btrue = \n')
    print(Btrue)
    cat('Bknow = \n')
    print(Bknow)
    cat('Ctrue = \n')
    print(Ctrue)
    cat('Cknow = \n')
    print(Cknow)
    cat('Btottrue = \n')
    print(Btottrue)
    cat('Btotknow = \n')
    print(Btotknow)
     
     if (any((Bknow == 0) && (Btrue != 0)) ||
         any((Bknow == 1) && (Btrue != 1)) ||
        any((Cknow == 0) && (Ctrue != 0)) ||     
        any((Cknow == 1) && (Ctrue != 1)) ||
        any((Btotknow == 0) && (Btottrue != 0)) ||
         any((Btotknow == 1) && (Btottrue != 1)))        
     {
       stop('An error occured! Should not happen, so something is seriously wrong.')
     }
     else
     {
       cat('Everything seems to be in order.\n')
     }
  } 

  #------------------------------------------------------------------------
  # Task 2: Runtime as a function of the number of variables
  #------------------------------------------------------------------------

  if (task==2) {

    # PARAMETERS

    # True model parameters
    nvec <- 5:9
    nexp <- 10
    pdir <- 0.2
    pconf <- 0.2
    nmodels <- 100 
        
    # We will not be making additional assumptions when these when doing the inference here
    assumeacyclic <- 0
    assumenoconfounders <- 0

    # INITIALIZATIONS
    
    med2 <- rep(0,length(nvec))
    medinf <- rep(0,length(nvec))
    maxcset2secs <- rep(0,nmodels)
    maxcsetinfsecs <- rep(0,nmodels)

    # DO THE INFERENCE FOR EACH CASE

    nind <- 0
    for (n in nvec)
    {      
      nind <- nind+1
      
      # Show progress
      cat('- Starting case n=',n,'\n',sep='')

      secs <- rep(0,nmodels)
      for (i in 1:nmodels)
      {
        # Show progress
        cat('  * Starting model',i,'of',nmodels,'\n')

        # Create Btrue, Ctrue, and E
        cat('    - Creating the true model\n')
        Btrue <- matrix( (runif(n*n)<pdir)*1, n, n )
        diag(Btrue) <- 0
        Ctrue <- matrix( (runif(n*n)<pconf)*1, n, n )
        Ctrue[lower.tri(Ctrue)] <- 0
        diag(Ctrue) <- 0
        Ctrue <- Ctrue + t(Ctrue)
        E <- matrix( floor(runif(n*nexp)*3), n, nexp )

        # Computing the total effects
        cat('    - Computing the true total effects\n')
        Btottrue <- matrix( 0, n, n )
        diag(Btottrue) <- 1
        for (dummy in 1:n) Btottrue <- (Btottrue | (Btrue %*% Btottrue))*1

        # Write the model to file
        cat('    - Writing the model to file\n')
        fname <- 'truemodelfile.txt'
        writemodelfile( fname, Btrue, Ctrue, E )

        # Call discoverer3 to do the inference
        cat('    - Running discoverer3 with maxcset=2\n')
        starttimer()      
        maxcset <- 2
        out2 <- system(paste('./discoverer3',fname,maxcset,
        assumeacyclic, assumenoconfounders,'>out2.txt',sep=' '))        
        if (out2)
        {
          stop('The discoverer3 program gave a non-zero return value!')
        }
        elapsed <- stoptimer()
        cat('    - Time spent:',elapsed,'seconds.\n')
        maxcset2secs[i] <- elapsed
        
        # Call discoverer3 to do the inference
        cat('    - Running discoverer3 with maxcset=Inf\n')
        starttimer()      
        maxcset <- -1
        outinf <- system(paste('./discoverer3',fname,maxcset,
        assumeacyclic, assumenoconfounders,'>outinf.txt',sep=' '))        
        if (outinf)
        {
          stop('The discoverer3 program gave a non-zero return value!')
        }
        elapsed <- stoptimer()
        cat('    - Time spent:',elapsed,'seconds.\n')
        maxcsetinfsecs[i] <- elapsed        
      }
      
      # compute min, max, and median and store it
      med2[nind] <- median(maxcset2secs)
      medinf[nind] <- median(maxcsetinfsecs)  
      
      cat('med2 = ')
      print(med2)
      cat('medinf = ')
      print(medinf)   
    }

    overallmin <- min(med2)
    overallmax <- max(medinf)
        
    # Create the plot
    n <- nvec
    seconds <- rep(overallmin,length(nvec))
    seconds[2] <- overallmax
    plot(n, seconds, type="n", log="y")
    lines(nvec, medinf, lty=1, col=3)
    lines(nvec, med2, lty=2, col=2)
        
  } 

  #-----------------------------------------------------------------------------------
  # Task 3: Affect of limiting conditioning set size on identifiability
  #-----------------------------------------------------------------------------------

  if (task==3) {

    # PARAMETERS

    # True model parameters
    n <- 8
    maxcsetvec <- 0:(n-2)
    nexp <- 10
    pdir <- 0.2
    pconf <- 0.2
    nmodels <- 100
        
    # We will not be assuming these when doing the inference
    assumeacyclic <- 0
    assumenoconfounders <- 0

    # THESE WILL STORE THE STATS

    Btruezeros <- 0
    Bidentzeros <- rep(0,length(maxcsetvec))
    Btrueones <- 0
    Bidentones <- rep(0,length(maxcsetvec))

    Ctruezeros <- 0
    Cidentzeros <- rep(0,length(maxcsetvec))
    Ctrueones <- 0
    Cidentones <- rep(0,length(maxcsetvec))
    
    Btottruezeros <- 0
    Btotidentzeros <- rep(0,length(maxcsetvec))
    Btottrueones <- 0
    Btotidentones <- rep(0,length(maxcsetvec))
            
    # CREATE THE MODELS TO BE USED

    cat('Creating the models.\n')
    models <- list()
    for (i in 1:nmodels)
    {      
      # Create Btrue, Ctrue, and E
      Btrue <- matrix( (runif(n*n)<pdir)*1, n, n )
      diag(Btrue) <- 0
      Ctrue <- matrix( (runif(n*n)<pconf)*1, n, n )
        Ctrue[lower.tri(Ctrue)] <- 0
      diag(Ctrue) <- 0
      Ctrue <- Ctrue + t(Ctrue)
      E <- matrix( floor(runif(n*nexp)*3), n, nexp )
        
      # Computing the total effects
      Btottrue <- matrix( 0, n, n )
      diag(Btottrue) <- 1
      for (dummy in 1:n) Btottrue <- (Btottrue | (Btrue %*% Btottrue))*1
    
      # Store it all
      thismodel <- list()
      thismodel$Btrue <- Btrue
      thismodel$Ctrue <- Ctrue
      thismodel$E <- E
      thismodel$Btottrue <- Btottrue
      models[[i]] <- thismodel
      
      # Update true model stats
      # (taking into account that diagonal is always 0 in Btrue and Ctrue
      # and always 1 in Btottrue, these should not count towards the results!)
      Btruezeros <- Btruezeros + (sum(Btrue == 0) - n)
      Btrueones <- Btrueones + sum(Btrue == 1)
      Ctruezeros <- Ctruezeros + (sum(Ctrue == 0) - n)/2
      Ctrueones <- Ctrueones + sum(Ctrue == 1)/2
      Btottruezeros <- Btottruezeros + sum(Btottrue == 0)
      Btottrueones <- Btottrueones + (sum(Btottrue == 1) - n)
            
    }

    # GO THROUGH EACH MODEL

    for (i in 1:nmodels)
    {    
      # Show progress
      cat('- Starting model',i,'of',nmodels,'\n')
      
      # Write the model to file
      fname <- 'truemodelfile.txt'
      writemodelfile( fname, models[[i]]$Btrue, models[[i]]$Ctrue, models[[i]]$E )

      # Call discoverer3 to do the inference
      maxcset <- -1
      out <- system(paste('./discoverer3',fname,maxcset,
                       assumeacyclic, assumenoconfounders,'>out.txt',sep=' '))        
      if (out)
      {
        # Some problem occurred. Inform the user
        stop('The discoverer3 program gave a non-zero return value!')
      }
      
      # Source the result file
      source('out.txt', local = TRUE)
      
      # Go through each maxcset value
      maxcsetind <- 0
      for (maxcset in maxcsetvec)
      {
        maxcsetind <- maxcsetind + 1
      
        # Get knowledge matrices corresponding to this maxcset value
        Bknow <- eval(parse(text = paste('Bknow',maxcset,sep='')))
        Cknow <- eval(parse(text = paste('Cknow',maxcset,sep='')))
        Btotknow <- eval(parse(text = paste('Btotknow',maxcset,sep='')))
                
        # Update the stats (again remember diagonal, see above)
        Bidentzeros[maxcsetind] <- Bidentzeros[maxcsetind] + (sum(Bknow == 0) - n)
        Bidentones[maxcsetind] <- Bidentones[maxcsetind] + sum(Bknow == 1)
        Cidentzeros[maxcsetind] <- Cidentzeros[maxcsetind] + (sum(Cknow == 0) - n)/2
        Cidentones[maxcsetind] <- Cidentones[maxcsetind] + sum(Cknow == 1)/2
        Btotidentzeros[maxcsetind] <- Btotidentzeros[maxcsetind] + sum(Btotknow == 0)
        Btotidentones[maxcsetind] <- Btotidentones[maxcsetind] + (sum(Btotknow == 1) - n)      
      }    
    }  
    
    
    # OUTPUT THE RESULTS IN TEXT FORM
    
    cat('Babsence = ')
    print(100*Bidentzeros/Btruezeros)
    cat('Bpresence = ')
    print(100*Bidentones/Btrueones)
    cat('Cabsence = ')
    print(100*Cidentzeros/Ctruezeros)
    cat('Cpresence = ')
    print(100*Cidentones/Ctrueones)
    cat('Btotabsence = ')
    print(100*Btotidentzeros/Btottruezeros)
    cat('Btotpresence = ')
    print(100*Btotidentones/Btottrueones)    
    
    # CREATE THE PLOTS
    
    overallmin <- 0
    overallmax <- 100
        
    # Create the plot
    maxcset <- maxcsetvec
    id <- rep(0,length(maxcsetvec))
    id[2] <- 100
    plot(maxcset, id, type="n")
    lines(maxcset, 100*Bidentzeros/Btruezeros, lty=2, col=2)
    lines(maxcset, 100*Bidentones/Btrueones, lty=1, col=2)
    lines(maxcset, 100*Cidentzeros/Ctruezeros, lty=2, col=3)
    lines(maxcset, 100*Cidentones/Ctrueones, lty=1, col=3)
    lines(maxcset, 100*Btotidentzeros/Btottruezeros, lty=2, col=4)
    lines(maxcset, 100*Btotidentones/Btottrueones, lty=1, col=4)
    
    
  } 


  #-----------------------------------------------------------------------------------
  # Task 4: How much more is learned with an assumption of acyclicity, or causal 
  # sufficiency, or both, in cases when these assumptions hold? 
  #-----------------------------------------------------------------------------------

  if (task==4) {

    # PARAMETERS

    # True model parameters
    n <- 8
    maxcset <- -1
    sparsity <- 20
    expmode <- 'norm'
    nmodels <- 100

    # GO THROUGH EACH CONDITION, CREATE THE MODELS, DO THE INFERENCE,
    # AND PERFORM THE ANALYSIS

    for (acyclic in c(0,1))
    {
      for (noconfounders in c(0,1))
      {
        if ((acyclic==0) && (noconfounders==0)) next

        for (assumeacyclic in c(0,1))
        {
          for (assumenoconfounders in c(0,1))
          {
            # Here we don't consider cases where an assumption is made that does not hold   
            if ((assumeacyclic == 1) && (acyclic == 0)) next
            if ((assumenoconfounders == 1) && (noconfounders == 0)) next
            
            # Show progress
            cat('Starting new condition: \n')
            if (acyclic) 
            {
              cat('ACYCLIC MODELS')
              if (assumeacyclic) cat(' - ASSUMING ACYCLICITY!')
              cat('\n')
            }
            if (noconfounders) 
            {
              cat('NO CONFOUNDER MODELS')
              if (assumenoconfounders) cat(' - ASSUMING NO CONFOUNDERS!')
              cat('\n')
            }
            
            # Initialize statistics
            directedges <- 0
            directedgestot <- 0
            directabsences <- 0
            directabsencestot <- 0
            totaledges <- 0
            totaledgestot <- 0
            totalabsences <- 0
            totalabsencestot <- 0
            confounderedges <- 0
            confounderedgestot <- 0
            confounderabsences <- 0
            confounderabsencestot <- 0
            
            # Go through the models
            for (themodel in 1:nmodels)
            {
              cat('[',themodel,']\n',sep='')
  
              # Set seed
              set.seed( themodel-1 )

              # Create Btrue, Ctrue, and E
              Btrue <- CreateBtrue(n, sparsity, acyclic)
              Ctrue <- CreateCtrue(n, sparsity, noconfounders)
              E <- CreateE(n, expmode)
              nexp <- ncol(E)   
            
              # Computing the total effects
              Btottrue <- matrix( 0, n, n )
              diag(Btottrue) <- 1
              for (dummy in 1:n) Btottrue <- (Btottrue | (Btrue %*% Btottrue))*1
              
              # Do the inference
              fname <- 'truemodelfile.txt'
              writemodelfile( fname, Btrue, Ctrue, E )

              # Call the C program and get the output
              cat('Calling discoverer.\n')
              out <- system(paste('./discoverer3',fname,maxcset,
              assumeacyclic, assumenoconfounders,'>out.txt',sep=' '))        
              if (out)
              {
                # Some problem occurred. Inform the user
                stop('The discoverer3 program gave a non-zero return value! Something went wrong!')
              }
              cat('Done! Resuming operation in R.\n')    

              # Source the file
              source('out.txt', local = TRUE)

              # Update statistics
              directedgestot <- directedgestot + sum(Btrue == 1)
              directabsencestot <- directabsencestot + (sum(Btrue == 0) - n)
              totaledgestot <- totaledgestot + (sum(Btottrue == 1) - n)
              totalabsencestot <- totalabsencestot + sum(Btottrue == 0)
              confounderedgestot <- confounderedgestot + (sum(Ctrue == 1)/2)
              confounderabsencestot <- confounderabsencestot + ((sum(Ctrue == 0) - n)/2)
              directedges <- directedges + sum(Bknow == 1)
              directabsences <- directabsences + (sum(Bknow == 0) - n)
              totaledges <- totaledges + (sum(Btotknow == 1) - n)
              totalabsences <- totalabsences + sum(Btotknow == 0)
              confounderedges <- confounderedges + (sum(Cknow == 1)/2)
              confounderabsences <- confounderabsences + ((sum(Cknow == 0) - n)/2)          

        
            }

            # Show the results
            cat('\n')
            if (acyclic) cat('True model is ACYCLIC\n') 
            if (noconfounders) cat('True model has NO CONFOUNDERS\n')
            if (assumeacyclic) cat('Assuming ACYCLIC\n')
            if (assumenoconfounders) cat('Assuming NO CONFOUNDERS\n')
            cat('Proportion of identified direct edges:          ',directedges/directedgestot,'\n')
            cat('Proportion of identified direct absences:       ',directabsences/directabsencestot,'\n')
            cat('Proportion of identified total-effect edges:    ',totaledges/totaledgestot,'\n')
            cat('Proportion of identified total-effect absences: ',totalabsences/totalabsencestot,'\n')
            cat('Proportion of identified confounder edges:      ',confounderedges/confounderedgestot,'\n')
            cat('Proportion of identified confounder absences:   ',confounderabsences/confounderabsencestot,'\n')

          }
        }
      }
    }
  }

  # Display the seed used, just in case something went wrong and we 
  # would like to start debugging.
  cat('Used the seed: ',theseed,'\n')

}
