
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "matrix.h"
#include "indeps.h"
#include "oracle.h"
#include "minisatbb3api.h"
#include "edgepathindex.h"

// Number of variables
static int n;

// Number of experiments
static int nexp;

// Number of edges in the model
static long long nBinVar;

// Storing the experiment matrix here
static matrix Emat;

// Matrix to go from index to edge
static matrix index2edgeTable;

// Matrices to go from edge to index
static matrix directedge2indexTable;
static matrix confounderedge2indexTable;

// Helper static vars
static long long nIndep;               // = (n*(n-1)/2)*(2^(n-2))
static long long nSubsets;             // = 2^n
static long long nactualpairs;         // = n*(n-1)/2
static long long npathpairs;           // = n*(n+1)/2
static long long npathtypes;           // = 4
static long long nunknowns;            // total number of SAT variables (unknowns)
static int maxpathlength;              // maximum path length to consider (e.g. 2*n)

static matrix vec;
static matrix vecsub;

// Holds which path constraints have been added
static matrix pathconstraintsadded;



// --- Functions to handle edges and paths ---

// The total number of edges is nBinVar = 3*(n*(n-1)/2), i.e. three possible edges between
// each (unordered) pair of variables. They will be assigned edge IDs as follows: First,
// we list all the directed edges of which there are 2*(n*(n-1)/2). They are listed
// in the order of columnwise scanning of the matrix B[to,from], that is, 1=>2, 1=>3,
// ... 1=>n, then 2=>1, 2=>3, ... 2=>n, and so on. Then come the confounders, listed
// in columnwise scanning of the C matrix, neglecting entries on and above the diagonal,
// i.e. {2,1}, {3,1}, ... {n,1}, {3,2}, {4,2}, ... {n,2}, {4,3}, {5,3}, ... and so on.
// The ordering does not really make any difference, we just need to make sure that the
// functions dealing directly with edge IDs (these are few) are consistent.

// The constructed edge indices are used as follows, where 'i' is the index (0 <= i < nBinVar):
// el(index2edgeTable, i, 0 ) = startNode
// el(index2edgeTable, i, 1 ) = stopNode
// el(index2edgeTable, i, 2 ) = {1 or 3} (1 for startNode => stopNode, 3 for confounder)
// el(directedge2indexTable, stopNode, startNode) = i
// el(confounderedge2indexTable, stopNode, startNode ) = i, or equivalently:
// el(confounderedge2indexTable, startNode, stopNode ) = i

void InitEdgeIndex( int nvars, matrix E )
{
  // Helper integers
  int i,startNode,stopNode;

  // Set the number of variables
  n = nvars;
  nexp = E.cols;
  
  // Store the experiment matrix
  Emat = CopyMatrix( E );
  
  // Set the number of edges
  nBinVar = 3*n*(n-1)/2;
  
  nIndep = (n*(n-1)/2) * ((long long) pow((double) 2,n-2));
  nSubsets = ((long long) pow((double) 2,n));
  nactualpairs = n*(n-1)/2;
  npathpairs = n*(n+1)/2;
  npathtypes = 4;
  
  vec = CreateMatrix(1, n, 0);
  vecsub = CreateMatrix(1, n, 0);
  pathconstraintsadded = CreateMatrix( nSubsets, nexp, 0);
  
  maxpathlength = (2*n)-4;
  if (maxpathlength < (n-1)) maxpathlength = n-1;

  // Number of SAT unknowns. Will be incremented while generating the constraints
  // when auxiliary variables are needed. This now takes into account the
  // sat variables that define existence of *directed paths in the true model*
  // that we use to build up the total effects.
  
  nunknowns = nBinVar + nexp*nIndep + nexp*nSubsets*maxpathlength*npathpairs*npathtypes +
              2*nactualpairs + maxpathlength*2*nactualpairs;
  

  
  // Initialize the index number
  i = -1;
  
  // Initialize tables
  index2edgeTable = CreateMatrix( nBinVar, 3, 0 );
  directedge2indexTable = CreateMatrix( n, n, 0 );
  confounderedge2indexTable = CreateMatrix( n, n, 0 );
  
  // Step through all directed edges
  for (startNode=0; startNode<n; startNode++)
  {
    for (stopNode=0; stopNode<n; stopNode++)
    {
      if (startNode == stopNode) continue;
      else
      {
        // Increment counter
        i = i+1;
        
        // Add to index2edgeTable
        el(index2edgeTable,i,0) = startNode;
        el(index2edgeTable,i,1) = stopNode;
        el(index2edgeTable,i,2) = 1;
        
        // Add to directedge2indexTable
        el(directedge2indexTable,stopNode,startNode) = i;
      }
    }
  }
  
  // Step through all confounding edges
  for (startNode=0; startNode<n; startNode++)
  {
    for (stopNode=startNode+1; stopNode<n; stopNode++)
    {
      // Increment counter
      i = i+1;
      
      // Add to index2edgeTable
      el(index2edgeTable,i,0) = startNode;
      el(index2edgeTable,i,1) = stopNode;
      el(index2edgeTable,i,2) = 3;
      
      // Add to confounderedge2indexTable
      el(confounderedge2indexTable,stopNode,startNode) = i;
      el(confounderedge2indexTable,startNode,stopNode) = i;
    }
  }
  
  // We're done
  return;
  
}

int GetMaxPathLength( void )
{
  return maxpathlength;
}

void PrintEdgeIndex( void )
{
  printf("nBinVar = %lld\n",nBinVar);
  PrintMatrix("index2edgeTable = ", index2edgeTable );
  PrintMatrix("directedge2indexTable = ", directedge2indexTable );
  PrintMatrix("confounderedge2indexTable = ", confounderedge2indexTable );
}

int DirectEdgeToIndex( int from, int to )
{
  return (el(directedge2indexTable, to, from));
}

int ConfounderEdgeToIndex( int from, int to )
{
  return (el(confounderedge2indexTable, to, from));
}

int GetIndexToEdgeTableElement( int i, int j )
{
  return (el(index2edgeTable, i, j));
}

//-----------------------------------------------------------------------//
// Functions for indexing the underlying edges
// (both in underlying model and in the experiments)
//-----------------------------------------------------------------------//

// Edge x->y in the underlying model, with x != y
long long GetOrigDirEdge(int x, int y)
{
  // Check that x is not equal to y, and both are in the right range.
  CheckXnotequalY( x, y );
  
  // Get the edge index, add one to make it base 1.
  return el(directedge2indexTable,y,x)+1;
}

// Confounder x<->y in the underlying model, with x != y
long long GetOrigConfEdge(int x, int y)
{
  // Check that x is not equal to y, and both are in the right range.
  CheckXnotequalY( x, y );
  
  // Get the confounder index, add one to make it base 1
  return el(confounderedge2indexTable,y,x)+1;
}

// Edge x->y in the experiment ei, with x != y
// NOTE: returns 0 in case the edge is cut by the experiment, or if x==y
long long GetExpDirEdge(int x, int y, int ei)
{
  // Check that ei is in the appropriate range
  Checkei( ei );
  
  if (x==y) return 0;
  
  // If the edge is cut by the experiment we return 0
  if (el(Emat,y,ei)==2)
  {
    return 0;
  }
  else
  {
    return GetOrigDirEdge(x, y);
  }
}

// Confounder x<->y in the experiment ei, with x != y
// NOTE: returns 0 in case the edge is cut by the experiment, or if x==y
long long GetExpConfEdge(int x, int y, int ei)
{
  // Check that ei is in the appropriate range
  Checkei( ei );
  
  if (x==y) return 0;
  
  // If the edge is cut by the experiment return 0
  if ((el(Emat,y,ei)==2) || (el(Emat,x,ei)==2))
  {
    return 0;
  }
  else
  {
    return GetOrigConfEdge(x, y);
  }
}


//-----------------------------------------------------------------------//
// Function for indexing the d-separation variables
// These come directly after the edge variables, and
// they follow the ordering of the possIndepList, with
// an outer loop for going through the experiments
//-----------------------------------------------------------------------//


long long GetExpDsep(long long ii, int ei)
{
  // Check that arguments are valid
  Checkii( ii );
  Checkei( ei );
  
  // Output the index
  long long satvar = 1 + nBinVar + (ei*nIndep) + ii;
  
  // Return the index
  return satvar;
}

//-----------------------------------------------------------------------//
// Function for indexing the "existence of paths" variables
//-----------------------------------------------------------------------//

// Active path from x to y (x<y) with conditioning set 'si' of type 'pathtype' in the experiment ei
// (Note: pathtype forward vs backward refers to the node with smaller index being 'first'
// and the node with larger index being 'later')
long long GetExpPath(int pathtype, int x, int y, int length, long long si, int ei)
{
  // Check that inputs are in the right range.
  Checksi( si );
  Checklength( length );
  Checkei( ei );
  
  // Check that x is less than (or equal to y), and both are in the right range.
  CheckXlessthanorequaltoY( x, y );
  
  // Get the pair index (base 0)
  int i = 0;
  int whichpair;
  for (int xi=0; xi<n; xi++)
  {
    for (int yi=xi; yi<n; yi++)
    {
      if ((x==xi) && (y==yi))
      {
        whichpair = i;
      }
      i++;
    }
  }
  
  // These are the ones prior to the path variables
  long long satvar = 1 + nBinVar + nexp*nIndep;
  
  // Now go to the appropriate one
  satvar = satvar + ei*nSubsets*maxpathlength*npathpairs*npathtypes +
           si*maxpathlength*npathpairs*npathtypes +
           length*npathpairs*npathtypes +
           whichpair*npathtypes +
           pathtype;
  
  // Return
  return satvar;
}

//-----------------------------------------------------------------------//
// Function for indexing the total effect variables
//-----------------------------------------------------------------------//

// Directed path from x to y in the true model
long long GetTotalEffect(int x, int y)
{
  // Check that x and y are in the right range and not equal.
  CheckXnotequalY( x, y );
  
  // Get the pair index (base 1)
  long long satvar = GetOrigDirEdge(x, y);
  
  // Add everything before it
  satvar = satvar + nBinVar + nexp*nIndep + nexp*nSubsets*maxpathlength*npathpairs*npathtypes;
  
  // Return
  return satvar;
}

//-----------------------------------------------------------------------//
// Function for indexing the paths building up to the total effect
//-----------------------------------------------------------------------//

// Directed path from x to y, with a given length, in the true model
// Note: length 0 is a direct effect.
long long GetTotalEffectPath(int x, int y, int length)
{
  // Check that x and y are in the right range and not equal.
  CheckXnotequalY( x, y );
  
  // Check that length is in the right range
  Checklength( length );
  
  // Get the pair index (base 1)
  long long satvar = GetOrigDirEdge(x, y);
  
  // Subtract 1 to get base 0
  satvar = satvar - 1;
  
  // Get the current one
  satvar = satvar*maxpathlength + length;
  
  // Add 1 to get to base 1
  satvar = satvar+1;
  
  // Add everything before it
  satvar = satvar + nBinVar + nexp*nIndep +
           nexp*nSubsets*maxpathlength*npathpairs*npathtypes +
           2*nactualpairs;
  
  // Return
  return satvar;
}



//-----------------------------------------------------------------------//
// Functions to check for errors (just in case!)
//-----------------------------------------------------------------------//

void CheckXnotequalY( int x, int y )
{
  if ((x < 0) || (x>=n))
  {
    printf("x out of range! exiting!");
    exit(EXIT_FAILURE);
  }
  
  if ((y < 0) || (y>=n))
  {
    printf("y out of range! exiting!");
    exit(EXIT_FAILURE);
  }
  
  if (x == y)
  {
    printf("x equal to y! exiting!");
    exit(EXIT_FAILURE);
  }
}

void CheckXlessthanY( int x, int y )
{
  if ((x < 0) || (x>=n))
  {
    printf("x out of range! exiting!");
    exit(EXIT_FAILURE);
  }
  
  if ((y < 0) || (y>=n))
  {
    printf("y out of range! exiting!");
    exit(EXIT_FAILURE);
  }
  
  if (x >= y)
  {
    printf("x larger than y! exiting!");
    exit(EXIT_FAILURE);
  }
}

void CheckXlessthanorequaltoY( int x, int y )
{
  if ((x < 0) || (x>=n))
  {
    printf("x out of range! exiting!");
    exit(EXIT_FAILURE);
  }
  
  if ((y < 0) || (y>=n))
  {
    printf("y out of range! exiting!");
    exit(EXIT_FAILURE);
  }
  
  if (x > y)
  {
    printf("x larger than y! exiting!");
    exit(EXIT_FAILURE);
  }
}

void Checkx( int x )
{
  if ((x < 0) || (x>=n))
  {
    printf("x out of range! exiting!");
    exit(EXIT_FAILURE);
  }
}

void Checkii( long long ii )
{
  if ((ii < 0) || (ii>=nIndep))
  {
    printf("ii out of range! exiting!");
    exit(EXIT_FAILURE);
  }
}

void Checksi( long long si )
{
  if ((si < 0) || (si>=nSubsets))
  {
    printf("si out of range! exiting!");
    exit(EXIT_FAILURE);
  }
}

void Checklength( int length )
{
  if ((length < 0) || (length>=maxpathlength))
  {
    printf("length out of range! exiting!");
    exit(EXIT_FAILURE);
  }
}

void Checkei( int ei )
{
  if ((ei < 0) || (ei>=nexp))
  {
    printf("ei out of range! exiting!");
    exit(EXIT_FAILURE);
  }
}

long long GetCondSet( long long ii )
{
  long long si = 0;
  long long fact = 1;
  
  // Get the vector representation
  matrix vec = CreateMatrix(1, n, 0);
  IndepIndexToIndepVec( ii, vec );
  
  // Compute the condset index
  for (int j=0; j<n; j++)
  {
    if (el(vec,0,j)==1) si += fact;
    fact *= 2;
  }

  // Return it
  return si;
}



void WriteTotalEffectConstraints( void )
{
  // Array to hold a single constraint
  long long vars[maxpathlength+1];
  
  // First we define a total effect in terms of existence of a directed
  // path, of any length, from a variable to the other.
  for (int x=0; x<n; x++)
  {
    for (int y=0; y<n; y++)
    {
      if (x==y) continue;
      
      // Get the total effect variable
      int T = GetTotalEffect(x, y);
      
      // Get all the total effect paths of various lengths
      long long pathind[maxpathlength];
      for (int length=0; length<maxpathlength; length++)
      {
        pathind[length] = GetTotalEffectPath(x, y, length);
      }
      
      // Write the constraints
      for (int i=0; i<maxpathlength; i++)
      {
        vars[0] = T;
        vars[1] = -pathind[i];
        satconstraint( 2, vars );
      }
      vars[0] = -T;
      for (int i=0; i<maxpathlength; i++) vars[i+1] = pathind[i];
      satconstraint( maxpathlength+1, vars );
    }
  }
  
  // Next we define the existence of length 0 paths
  for (int x=0; x<n; x++)
  {
    for (int y=0; y<n; y++)
    {
      if (x==y) continue;
      
      // Get the total effect path variable
      long long T = GetTotalEffectPath(x, y, 0);
      
      // Get the direct effect variable
      long long d = GetOrigDirEdge(x, y);
      
      // Write the constraints
      vars[0] = T;
      vars[1] = -d;
      satconstraint( 2, vars );
      vars[0] = -T;
      vars[1] = d;
      satconstraint( 2, vars );
    }
  }
  
  // Finally we define the existence of length l path
  for (int x=0; x<n; x++)
  {
    for (int y=0; y<n; y++)
    {
      if (x==y) continue;
      
      for (int length=1; length<maxpathlength; length++)
      {
        // Get the total effect path variable
        long long T = GetTotalEffectPath(x, y, length);
        
        // Construct the array to hold the auxiliary variables
        int naux = 0;
        long long auxind[n];
        
        // Go through all other variables
        for (int z=0; z<n; z++)
        {
          if (z == x) continue;
          if (z == y) continue;
          
          long long Ek = GetOrigDirEdge(x, z);
          
          // Get a new unknown, use it for this conjunction
          nunknowns++;
          long long Ai = nunknowns;
          auxind[naux] = Ai;
          naux++;
          
          int P1 = GetTotalEffectPath(z, y, length-1);
          
          // Write out the constraints
          vars[0] = T;
          vars[1] = -Ai;
          satconstraint( 2, vars );
          vars[0] = -Ai;
          vars[1] = Ek;
          satconstraint( 2, vars );
          vars[0] = -Ai;
          vars[1] = P1;
          satconstraint( 2, vars );
          vars[0] = Ai;
          vars[1] = -Ek;
          vars[2] = -P1;
          satconstraint( 3, vars );
        }
        
        // Write out the final constraint that defines the OR
        vars[0] = -T;
        for (int i=0; i<naux; i++) vars[i+1] = auxind[i];
        satconstraint( naux+1, vars );
      }
    }
  }
}

void WriteAcyclicityConstraints( void )
{
  // Array to hold a single constraint
  long long vars[2];

  for (int x=0; x<n; x++)
  {
    for (int y=x+1; y<n; y++)
    {
      // Get the total effect variables
      long long T1 = GetTotalEffect(x, y);
      long long T2 = GetTotalEffect(y, x);
            
      // Write the constraint
      vars[0] = -T1;
      vars[1] = -T2;
      satconstraint( 2, vars );
    }
  }
}

void WriteNoConfoundersConstraints( void )
{
  // Array to hold a single constraint
  long long vars[2];

  for (int x=0; x<n; x++)
  {
    for (int y=x+1; y<n; y++)
    {
      // Get the sat variable
      long long satvar = GetOrigConfEdge(x, y);
      
      // Write the constraint
      vars[0] = -satvar;
      satconstraint( 1, vars );
    }
  }
}

void WriteSpecificDseps( long long ii, int ei, int isindep )
{
  // Array to hold a single constraint
  long long vars[2];

  // Get the satvar for this
  long long satvar = GetExpDsep(ii, ei);
  
  if (isindep == 1)
  {
    // Independence
    vars[0] = -satvar;
    satconstraint( 1, vars );
  }
  else if (isindep == 0)
  {
    // Dependence
    vars[0] = satvar;
    satconstraint( 1, vars );
  }
  else
  {
    printf("Should not try to add non-binary dsep value!\n");
    exit(EXIT_FAILURE);
  }
}

void WriteSpecificConstraints( long long ii, int ei )
{ 
  // Array to hold a single constraint
  long long vars[maxpathlength*npathtypes+1];
  
  // First we define the d-separations in terms of existence of paths
  // dsep(ii, ei) iff
  //  any(path(pathtype,x,y,length,si,ei))
  //  where pathtype = (ForwardPath, BackwardPath, BidirPath, UndirPath)
  //  and length = (0, ..., maxpathlength-1)
  //  and x<y are the endpoints,
  //  and si is the subset conditioning on everything required in ii *but not x nor y*
  //  (in principle, should be no difference if x or y are included, but we for simplicity
  //  just leave these out)
  //
  // Which turns into the CNF form
  //
  // D -P1
  //  ...
  // D -Pk
  // -D P1 ... Pk
  
  // Get the dsep variable index
  long long D = GetExpDsep(ii, ei);
  
  // Get the vector defining the possible independence
  IndepIndexToIndepVec( ii, vec );
  
  // Get the endpoints
  int x = -1;
  int y = -1;
  for (int j=0; j<n; j++)
  {
    if (el(vec,0,j)==2)
    {
      if (x<0) x=j;
      else y=j;
    }
  }
  
  // Get the conditioning set
  long long si = GetCondSet( ii );
  SubIndexToSubVec( si, vecsub );
  
  // Get the path indices
  long long pathind[maxpathlength*4];
  int i=0;
  for (int length=0; length<maxpathlength; length++)
  {
    for (int pathtype=0; pathtype<npathtypes; pathtype++)
    {
      pathind[i] = GetExpPath(pathtype, x, y, length, si, ei);
      i++;
    }
  }
  
  // Write the constraints
  for (i=0; i<maxpathlength*npathtypes; i++)
  {
    vars[0] = D;
    vars[1] = -pathind[i];
    satconstraint( 2, vars );
  }
  vars[0] = -D;
  for (i=0; i<maxpathlength*npathtypes; i++)
  {
    vars[i+1] = pathind[i];
  }
  satconstraint( maxpathlength*npathtypes + 1, vars );

  // Next we define the existence of paths in terms of edges and paths:
  
  // CNF form for constraint P iff ((A1 and B1) OR (A2 and B2) OR ... (Ak and Bk))
  // 1 auxiliary variable for each AND term, call these C1 ... Ck
  // For each i=1...k:
  // P -Ci
  // -Ci Ai
  // -Ci Bi
  // Ci -Ai -Bi
  // And add at the end:
  // -P C1 ... Ck
  //
  
  // If this combination of si and ei is already taken care of, then we can
  // skip this step. Otherwise we do it (and mark it done).
  if (el(pathconstraintsadded,si,ei)==0)
  {
    el(pathconstraintsadded,si,ei) = 1;
    
    for (int x=0; x<n; x++)
    {
      for (int y=x; y<n; y++)
      {
        
        // First handle length 0 paths. These are just equal to the underlying edges.
        // If the experiment breaks an edge then we should set the given path to FALSE
        // for this experiment.
        {
          long long P, E;
          
          P = GetExpPath(ForwardPath, x, y, 0, si, ei);
          E = GetExpDirEdge( x, y, ei );
          if (E)
          {
            vars[0] = P;
            vars[1] = -E;
            satconstraint( 2, vars );
            vars[0] = -P;
            vars[1] = E;
            satconstraint( 2, vars );
          }
          else
          {
            vars[0] = -P;
            satconstraint( 1, vars );
          }
          
          P = GetExpPath(BackwardPath, x, y, 0, si, ei);
          E = GetExpDirEdge( y, x, ei );
          if (E)
          {
            vars[0] = P;
            vars[1] = -E;
            satconstraint( 2, vars );
            vars[0] = -P;
            vars[1] = E;
            satconstraint( 2, vars );
          }
          else
          {
            vars[0] = -P;
            satconstraint( 1, vars );
          }
          
          P = GetExpPath(BidirPath, x, y, 0, si, ei);
          E = GetExpConfEdge( x, y, ei );
          if (E)
          {
            vars[0] = P;
            vars[1] = -E;
            satconstraint( 2, vars );
            vars[0] = -P;
            vars[1] = E;
            satconstraint( 2, vars );
          }
          else
          {
            vars[0] = -P;
            satconstraint( 1, vars );
          }
          
          P = GetExpPath(UndirPath, x, y, 0, si, ei);
          vars[0] = -P;
          satconstraint( 1, vars );
        }
                
        // Go through all possible lengths >= 1
        for (int length=1; length<maxpathlength; length++)
        {
          {
            // path(ForwardPath,[x,y,condset],length,ei) iff
            //  OR over all the following:
            //     for all z not equal to x nor y, with z not in condset:
            //       (directed edge x->z) and path(ForwardPath,[z,y,condset],length-1,ei)
            //     for all z not equal to x nor y, with z in condset:
            //       (directed edge x->z) and path(BidirPath,[z,y,condset],length-1,ei)
            
            long long P = GetExpPath(ForwardPath, x, y, length, si, ei);
            
            // Construct the array to hold the auxiliary variables
            int naux = 0;
            long long auxind[4*n];
            
            // Go through all z not in the conditioning set
            for (int z=0; z<n; z++)
            {
              if (el(vecsub,0,z)==1) continue;
              long long Ek = GetExpDirEdge( x, z, ei);
              if (Ek)
              {
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                long long Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P1;
                if (z < y) P1 = GetExpPath(ForwardPath, z, y, length-1, si, ei);
                else P1 = GetExpPath(BackwardPath, y, z, length-1, si, ei);
                
                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P1;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P1;
                satconstraint( 3, vars );
              }
            }
            
            // Go through all z with z in conditioning set
            for (int z=0; z<n; z++)
            {
              if (el(vecsub,0,z)!=1) continue;
              long long Ek = GetExpDirEdge( x, z, ei);
              if (Ek)
              {
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                long long Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P1;
                if (z < y) P1 = GetExpPath(BidirPath, z, y, length-1, si, ei);
                else P1 = GetExpPath(BidirPath, y, z, length-1, si, ei);
                
                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P1;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P1;
                satconstraint( 3, vars );
              }
            }
            
            // Write out the final constraint that defines the OR
            vars[0] = -P;
            for (int i=0; i<naux; i++) vars[i+1] = auxind[i];
            satconstraint( naux+1, vars );
          }
          
          {
            // path(UndirPath,[x,y,condset],length,ei) iff
            //  OR over all the following:
            //     for all z not equal to x nor y, with z not in condset:
            //       (directed edge x->z) and path(UndirPath,[z,y,condset],length-1,ei)
            //     for all z not equal to x nor y, with z in condset:
            //       (directed edge x->z) and path(BackwardPath,[z,y,condset],length-1,ei)
            
            long long P = GetExpPath(UndirPath, x, y, length, si, ei);
            
            // Construct the array to hold the auxiliary variables
            int naux = 0;
            long long auxind[4*n];
            
            // Go through all z not in the conditioning set
            for (int z=0; z<n; z++)
            {
              if (el(vecsub,0,z)==1) continue;
              long long Ek = GetExpDirEdge( x, z, ei);
              if (Ek)
              {
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                long long Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P1;
                if (z < y) P1 = GetExpPath(UndirPath, z, y, length-1, si, ei);
                else P1 = GetExpPath(UndirPath, y, z, length-1, si, ei);
                
                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P1;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P1;
                satconstraint( 3, vars );
              }
            }
            
            // Go through all z with z in conditioning set
            for (int z=0; z<n; z++)
            {
              if (el(vecsub,0,z)!=1) continue;
              long long Ek = GetExpDirEdge( x, z, ei);
              if (Ek)
              {
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                long long Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P1;
                if (z < y) P1 = GetExpPath(BackwardPath, z, y, length-1, si, ei);
                else P1 = GetExpPath(ForwardPath, y, z, length-1, si, ei);

                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P1;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P1;
                satconstraint( 3, vars );
              }
            }
            
            // Write out the final constraint that defines the OR
            vars[0] = -P;
            for (int i=0; i<naux; i++) vars[i+1] = auxind[i];
            satconstraint( naux+1, vars );
          }
          
          {
            // path(BidirPath,[x,y,condset],length,ei) iff
            //  OR over all the following:
            //     for all z not equal to x nor y, with z not in condset:
            //       (directed edge x<-z) and path(ForwardPath,[z,y,condset],length-1,ei)
            //       (directed edge x<-z) and path(BidirPath,[z,y,condset],length-1,ei)
            //       (confounder edge x<->z) and path(ForwardPath,[z,y,condset],length-1,ei)
            //     for all z not equal to x nor y, with z in condset:
            //       (confounder edge x<->z) and path(BidirPath,[z,y,condset],length-1,ei)
            
            long long P = GetExpPath(BidirPath, x, y, length, si, ei);
            
            // Construct the array to hold the auxiliary variables
            int naux = 0;
            long long auxind[4*n];
            
            // Go through all z not in the conditioning set
            for (int z=0; z<n; z++)
            {
              if (el(vecsub,0,z)==1) continue;
              long long Ek = GetExpDirEdge( z, x, ei);
              if (Ek)
              {
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                long long Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P1;
                if (z < y) P1 = GetExpPath(ForwardPath, z, y, length-1, si, ei);
                else P1 = GetExpPath(BackwardPath, y, z, length-1, si, ei);
                
                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P1;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P1;
                satconstraint( 3, vars );
                
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P2;
                if (z < y) P2 = GetExpPath(BidirPath, z, y, length-1, si, ei);
                else P2 = GetExpPath(BidirPath, y, z, length-1, si, ei);

                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P2;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P2;
                satconstraint( 3, vars );
              }
              
              Ek = GetExpConfEdge( x, z, ei);
              if (Ek)
              {
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                long long Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P1;
                if (z < y) P1 = GetExpPath(ForwardPath, z, y, length-1, si, ei);
                else P1 = GetExpPath(BackwardPath, y, z, length-1, si, ei);
                
                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P1;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P1;
                satconstraint( 3, vars );
              }
            }
            
            // Go through all z with z in conditioning set
            for (int z=0; z<n; z++)
            {
              if (el(vecsub,0,z)!=1) continue;
              long long Ek = GetExpConfEdge( x, z, ei);
              if (Ek)
              {
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                long long Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P1;
                if (z < y) P1 = GetExpPath(BidirPath, z, y, length-1, si, ei);
                else P1 = GetExpPath(BidirPath, y, z, length-1, si, ei);
                                
                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P1;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P1;
                satconstraint( 3, vars );
              }
            }
            
            // Write out the final constraint that defines the OR
            vars[0] = -P;
            for (int i=0; i<naux; i++) vars[i+1] = auxind[i];
            satconstraint( naux+1, vars );
          }
          
          {
            // path(BackwardPath,[x,y,condset],length,ei) iff
            //  OR over all the following:
            //     for all z not equal to x nor y, with z not in condset:
            //       (directed edge x<-z) and path(BackwardPath,[z,y,condset],length-1,ei)
            //       (directed edge x<-z) and path(UndirPath,[z,y,condset],length-1,ei)
            //       (confounder edge x<->z) and path(UndirPath,[z,y,condset],length-1,ei)
            //     for all z not equal to x nor y, with z in condset:
            //       (confounder edge x<->z) and path(BackwardPath,[z,y,condset],length-1,ei)
            
            long long P = GetExpPath(BackwardPath, x, y, length, si, ei);
            
            // Construct the array to hold the auxiliary variables
            int naux = 0;
            long long auxind[4*n];
            
            // Go through all z not in the conditioning set
            for (int z=0; z<n; z++)
            {
              if (el(vecsub,0,z)==1) continue;
              long long Ek = GetExpDirEdge( z, x, ei);
              if (Ek)
              {
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                long long Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P1;
                if (z < y) P1 = GetExpPath(BackwardPath, z, y, length-1, si, ei);
                else P1 = GetExpPath(ForwardPath, y, z, length-1, si, ei);
                
                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P1;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P1;
                satconstraint( 3, vars );
                                
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P2;
                if (z < y) P2 = GetExpPath(UndirPath, z, y, length-1, si, ei);
                else P2 = GetExpPath(UndirPath, y, z, length-1, si, ei);

                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P2;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P2;
                satconstraint( 3, vars );
              }
              
              Ek = GetExpConfEdge( x, z, ei);
              if (Ek)
              {
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                long long Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P1;
                if (z < y) P1 = GetExpPath(UndirPath, z, y, length-1, si, ei);
                else P1 = GetExpPath(UndirPath, y, z, length-1, si, ei);

                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P1;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P1;
                satconstraint( 3, vars );
              }
            }
            
            // Go through all z with z in conditioning set
            for (int z=0; z<n; z++)
            {
              if (el(vecsub,0,z)!=1) continue;
              long long Ek = GetExpConfEdge( x, z, ei);
              if (Ek)
              {
                // Get a new unknown, use it for this conjunction
                nunknowns++;
                long long Ai = nunknowns;
                auxind[naux] = Ai;
                naux++;
                
                // Get the path
                long long P1;
                if (z < y) P1 = GetExpPath(BackwardPath, z, y, length-1, si, ei);
                else P1 = GetExpPath(ForwardPath, y, z, length-1, si, ei);
                
                // Write out the constraints
                vars[0] = P;
                vars[1] = -Ai;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = Ek;
                satconstraint( 2, vars );
                vars[0] = -Ai;
                vars[1] = P1;
                satconstraint( 2, vars );
                vars[0] = Ai;
                vars[1] = -Ek;
                vars[2] = -P1;
                satconstraint( 3, vars );
              }
            }

            // Write out the final constraint that defines the OR
            vars[0] = -P;
            for (int i=0; i<naux; i++) vars[i+1] = auxind[i];
            satconstraint( naux+1, vars );
          }
        }
      }
    }
  }
}





