#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "matrix.h"
#include "indeps.h"
#include "oracle.h"
#include "edgepathindex.h"
#include "inference.h"
#include "minisatbb3api.h"
#include "utils.h"

static matrix Bknow, Cknow, Btotknow, Ematrix;
static int n, nexp, maxpathlength;
static matrix minIndep, maxIndep, dsepstatus, vec;
static long long nindep;

void InferModel(matrix Emat, int maxcondsetsize, int acyclicityConstraint, int noconfoundersConstraint)
{
  // Initializations
  Ematrix = CopyMatrix(Emat);
  n = Emat.rows;
  nexp = Emat.cols;
  maxpathlength = GetMaxPathLength();
  
  // Initialize the knowledge matrices
  Bknow = CreateMatrix(n, n, 2);
  Cknow = CreateMatrix(n, n, 2);
  Btotknow = CreateMatrix(n, n, 2);
  for (int i=0; i<n; i++)
  {
    el(Bknow,i,i)=0;
    el(Cknow,i,i)=0;
    el(Btotknow,i,i)=1;
  }
    
  // Helper vector
  vec = CreateMatrix(1, n, 0);
  
  // Initialize the minimum and maximum independencies matrices
  nindep = nTotIndep();
  minIndep = CreateMatrix(nindep, nexp, 3);
  maxIndep = CreateMatrix(nindep, nexp, 3);
  dsepstatus = CreateMatrix(nindep, nexp, 3);
  
  // Fire up the minisatbb3 SAT solver
  satinit( "tree" );
  
  // First, we load constraints defining the total effects, and any
  // acyclicity or noconfounder constraints
  {
    // Begin the LOAD operation
    satload();
    
    // Total effect constraints
    {
      printf("# - Writing the total effect constraints...\n");
      WriteTotalEffectConstraints();
      printf("#   Done!\n");
    }
    
    // Acyclicity constraints if requested
    if (acyclicityConstraint)
    {
      printf("# - Writing the acyclicity constraints...\n");
      WriteAcyclicityConstraints();
      printf("#   Done!\n");
    }
    
    // Noconfounders constraints if requested
    if (noconfoundersConstraint)
    {
      printf("# - Writing the noconfounders constraints...\n");
      WriteNoConfoundersConstraints();
      printf("#   Done!\n");
      
      // Set Cknow matrix to all zeros
      for (int x=0; x<n; x++)
      {
        for (int y=0; y<n; y++)
        {
          el(Cknow,x,y) = 0;
        }
      }
    }
    
    // End the LOAD operation
    satloadend();
  }
  
  // Solve one conditioning-set size at a time, but only adding those dseps that are
  // still unknown.
  for (int csetsize = 0; csetsize <= maxcondsetsize; csetsize++)
  {
    printf("# Starting iteration with csetsize = %i \n", csetsize);
    
    // Compute indep/dep/unknown status of all dseps with conditioning set size csetsize
    printf("# - Computing status of all dseps with this csetsize...\n");
    ComputeDsepStatus( csetsize );
    printf("#   Done!\n");
    
    // Begin the LOAD operation
    satload();
    
    // Go through all dseps
    printf("# - Adding constraints for any observable but unknown dseps...\n");
    long long totdsepsconsidered = 0;
    long long totdsepsunknown = 0;
    long long totdsepsadded = 0;
    for (long long ii=0; ii<nindep; ii++)
    {
      // Check if the given dsep has the appropriate size
      // of the conditioning set. If not, simply move to the next one.
      if (IndepCondSetSize( ii ) != csetsize) continue;
      
      // Go through all experiments
      for (int ei=0; ei<nexp; ei++)
      {
        // Construct manipulated versions of Bknow and Cknow
        matrix manBknow = CopyMatrix(Bknow);
        matrix manCknow = CopyMatrix(Cknow);
        for (int i=0; i<n; i++)
        {
          // Is variable i is intervened on?
          if (el(Emat,i,ei)==2)
          {
            // Remove any direct edges *into* i
            for (int j=0; j<n; j++) el(manBknow, i, j) = 0;
            
            // Remove any confounding edges between i and any other variable
            for (int j=0; j<n; j++)
            {
              el(manCknow, i, j) = 0;
              el(manCknow, j, i) = 0;
            }
          }
        }
        
        // Is the dsep observable?
        if (DSepOracleIndex( ii, ei ) != 3)
        {
          totdsepsconsidered++;
          
          // Only if the dsep is unknown we add it to the constraint set
          if (el(dsepstatus,ii,ei) == 2)
          {
            
            totdsepsunknown++;
            
            // Final check: is each element of the conditioning set potentially
            // connected to both endpoints (not through the other)?
            
            // This flag will be set to 0 if we found a problem.
            int checkok = 1;
    
            // Get the vector defining the possible independence
            IndepIndexToIndepVec( ii, vec );
            
            // Get the endpoints
            int x = -1;
            int y = -1;
            for (int j=0; j<n; j++)
            {
              if (el(vec,0,j)==2)
              {
                if (x<0) x=j;
                else y=j;
              }
            }

            
            // For each conditioned variable, check that it is ok.
            for (int z=0; z<n; z++)
            {
              if (el(vec,0,z)==1)
              {
                if (zInBetween(manBknow, manCknow, x, y, z) == 0)
                {
                  checkok = 0;
                }
              }
            }
            
            
            if (checkok)
            {
              totdsepsadded++;
              
              // Write constraints for this specific dsep
              WriteSpecificConstraints( ii, ei );
              WriteSpecificDseps( ii, ei, DSepOracleIndex(ii, ei) );
              
            }
          }
        }
      }
    }
    printf("#   Done.\n");
    printf("#   Total dseps:    %lld \n",totdsepsconsidered);
    printf("#   Unknown dseps:  %lld \n",totdsepsunknown);
    printf("#   Added dseps:    %lld \n",totdsepsadded);
    
    // End the LOAD operation
    satloadend();
        
    // Solve to get new knowledge generated from this csetsize.
    // (If no added dseps then nothing to do here!)
    if (totdsepsadded>0)
    {
      printf("# - Solving to update our model knowledge...\n");
  
      // Create the array of variables to handle
      long long vars[5*n*(n-1)/2];
      long long varstatus[5*n*(n-1)/2];
      int nvars = 0;
      
      // Add all variables that are unknown at this stage: B
      for (int x=0; x<n; x++)
      {
        for (int y=0; y<n; y++)
        {
          // Anything on the diagonal is by definition known
          if (x==y) continue;
          
          if (el(Bknow,y,x) == 2)
          {
            long long ind = GetOrigDirEdge(x, y);
            vars[ nvars ] = ind;
            nvars++;
          }
        }
      }

      // Add all variables that are unknown at this stage: C
      for (int x=0; x<n; x++)
      {
        for (int y=0; y<n; y++)
        {
          // Anything on the diagonal is by definition known
          if (x==y) continue;

          // For C we only consider the case x < y
          if (y<x) continue;
          
          if (el(Cknow,y,x) == 2)
          {
            long long ind = GetOrigConfEdge(x, y);
            vars[ nvars ] = ind;
            nvars++;
          }
        }
      }

      // Add all variables that are unknown at this stage: Btot
      for (int x=0; x<n; x++)
      {
        for (int y=0; y<n; y++)
        {
          // Anything on the diagonal is by definition known
          if (x==y) continue;
          
          if (el(Btotknow,y,x) == 2)
          {
            long long ind = GetTotalEffect(x, y);
            vars[ nvars ] = ind;
            nvars++;
          }
        }
      }
        
      // Copy the variable array to varstatus
      for (int i=0; i<nvars; i++) varstatus[i] = vars[i];
      
      // Ask for the backbone
      printf("# - Computing backbone...\n");
      int res = satbackbone( nvars, varstatus );
      printf("#   Done.\n");
      if (res == 0)
      {
        printf("We got UNSAT!? Aborting.\n");
        exit(EXIT_FAILURE);
      }
      if (res != 1)
      {
        printf("An error occured while computing the backbone.\n");
        exit(EXIT_FAILURE);
      }
              
      // Read the result file and get model knowledge
      printf("# - Reading current model knowledge...\n");
      GetModelKnowledge( nvars, vars, varstatus );
      printf("#   Done.\n");
      
    }
    
    // Show current model knowledge
    char text[100];
    sprintf(text,"Bknow%d = ",csetsize);
    PrintMatrix(text, Bknow);
    sprintf(text,"Cknow%d = ",csetsize);
    PrintMatrix(text, Cknow);
    sprintf(text,"Btotknow%d = ",csetsize);
    PrintMatrix(text, Btotknow);
    
  }
  
  // Quit minisatbb3
  satquit();
  
  // Show the result
  PrintMatrix("Bknow = ", Bknow);
  PrintMatrix("Cknow = ", Cknow);
  PrintMatrix("Btotknow = ", Btotknow);
    
  return;
}


void ComputeDsepStatus( int csetsize )
{
  // Creating minimal and maximal models from Bknow, Cknow
  matrix Bknowmin = CopyMatrix( Bknow );
  matrix Bknowmax = CopyMatrix( Bknow );
  for (int i=0; i<n; i++)
  {
    for (int j=0; j<n; j++)
    {
      if (el(Bknow,i,j) == 2)
      {
        el(Bknowmin,i,j) = 0;
        el(Bknowmax,i,j) = 1;
      }
    }
  }
  matrix Cknowmin = CopyMatrix( Cknow );
  matrix Cknowmax = CopyMatrix( Cknow );
  for (int i=0; i<n; i++)
  {
    for (int j=0; j<n; j++)
    {
      if (el(Cknow,i,j) == 2)
      {
        el(Cknowmin,i,j) = 0;
        el(Cknowmax,i,j) = 1;
      }
    }
  }
  
  // Get the minimal and maximal dseps
  ComputeDSeps( Bknowmin, Cknowmin, Ematrix, minIndep, csetsize, csetsize );
  ComputeDSeps( Bknowmax, Cknowmax, Ematrix, maxIndep, csetsize, csetsize );
  
  // Go through all, compute dsepstatus
  for (long long ii=0; ii<nindep; ii++)
  {
    // Check if the given dsep has the appropriate size
    // of the conditioning set. If not, simply move to the next one.
    if (IndepCondSetSize( ii ) != csetsize) continue;
    
    // Go through all experiments
    for (int ei=0; ei<nexp; ei++)
    {
      // Is the dsep unknown
      if (el(minIndep,ii,ei) != el(maxIndep, ii, ei))
      {
        el(dsepstatus, ii, ei) = 2;
      }
      else
      {
        el(dsepstatus, ii, ei) = el(minIndep,ii,ei);
      }
    }
  }
  
}

void GetModelKnowledge( int nvars, long long * vars, long long * varstatus )
{
  for (int i=0; i<nvars; i++)
  {
    long long thevar = vars[i];
    long long thevarstatus = varstatus[i];
    
    // Is it smaller than or equal to 3*n*(n-1)/2. If so, it is one of the
    // original edges in the model. If not, it is a total effect variable.
    if (thevar <= 3*n*(n-1)/2)
    {
      int from, to, type=0;
      
      // It is an edge in the underlying model. Find it.
      for (int x = 0; x < n; x++)
      {
        for (int y = 0; y < n; y++)
        {
          if (x == y) continue;
          if (GetOrigDirEdge( x, y ) == thevar)
          {
            from = x;
            to = y;
            type = 1;
          }
          else if (GetOrigConfEdge( x, y ) == thevar)
          {
            from = x;
            to = y;
            type = 3;
          }
        }
      }
      
      // Is it a a direct effect?
      if (type == 1)
      {
        if (thevarstatus == 1) el(Bknow,to,from) = 1;
        else if (thevarstatus == 0) el(Bknow,to,from) = 0;
        else if (thevarstatus == -1) el(Bknow,to,from) = 2;
      }
      // or is it a confounder edge?
      else if (type == 3)
      {
        if (thevarstatus == 1)
        {
          el(Cknow,to,from) = 1;
          el(Cknow,from,to) = 1;
        }
        else if (thevarstatus == 0)
        {
          el(Cknow,to,from) = 0;
          el(Cknow,from,to) = 0;
        }
        else if (thevarstatus == -1)
        {
          el(Cknow,to,from) = 2;
          el(Cknow,from,to) = 2;
        }
      }
      else
      {
        printf("No such edge type. Error.\n");
        printf("type = %d\n",type);
        exit(EXIT_FAILURE);
      }
    }
    else
    {
      // It is a total effect variable. Find it and set it
      for (int from = 0; from < n; from++)
      {
        for (int to = 0; to < n; to++)
        {
          if (from == to) continue;
          if (GetTotalEffect( from, to ) == thevar)
          {
            if (thevarstatus == 1) el(Btotknow,to,from) = 1;
            else if (thevarstatus == 0) el(Btotknow,to,from) = 0;
            else el(Btotknow,to,from) = 2;
          }
        }
      }
    }
  }
}

