#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <fcntl.h>

#include "minisatbb3api.h"
#include "matrix.h"
#include "utils.h"
#include "edgepathindex.h"
#include "indeps.h"
#include "oracle.h"
#include "inference.h"

int main(int argc, const char *argv[])
{   
  //------------------------------------------------------------------------------
  // SYNTAX for calling discoverer3:
  //
  // discoverer3 filename [maxcondsetsize [acyclic [noconfounders]]]
  // discoverer3 n nexp pdir pconf seed rformat [maxcondsetsize [acyclic [noconfounders]]]
  //
  // Where:
  //
  // filename           name of file containing model specification (from R)
  // n                  number of variables
  // nexp               number of experiments
  // pdir               probability of edge in B matrix
  // pconf              probability of edge in C matrix
  // seed               random number seed (-1 for clock-based varying seed)
  // rformat            1 for output suitable for reading into R, 0 otherwise
  // acyclic            1 for requiring acyclicity. default: 0
  // noconfounders      1 for assuming C is all zeros. default: 0
  // maxcondsetsize     maximum size of conditioning set (-1, ... , n-2). if
  //                    set to -1 it means n-2 (i.e. no limit). Default: -1
  //
  //------------------------------------------------------------------------------

  // Basic model parameters
  int n, nexp;
  matrix Btrue, Ctrue, Emat;

  // Random number seed used (for reproducibility)
  int randSeed;
  
  // Options which are read from the commandline arguments
  int maxcondsetsize;
  int acyclicityConstraint;
  int noconfoundersConstraint;
  
  printf("# Starting discoverer!\n");
  
  //------------------------------------------------------------------------------
  // Read the arguments, and load the model file or create a random model
  //------------------------------------------------------------------------------
  
  // Handle the case where a filename is given, specifying a file with the
  // model to use. This is the case when the number of arguments is 1-4 (so
  // argc is 2-5).
  
  if ((argc <= 5) && (argc >= 2))
  {
    // Set rformat to 1 always when reading the model from a file.
    SetRformat( 1 );
    
    // Initialize random number generator with seed 0
    randSeed = 0;
    InitRandWithSeed( randSeed );
    
    // Get and set the options
    // Note: If maxcondsetsize == -1 this is taken to mean no
    // constraint on conditioning set size.
    maxcondsetsize = ((argc > 2) ? ((int) atoi(argv[2])) : (-1));
    acyclicityConstraint = ((argc > 3) ? ((int) atoi(argv[3])) : (0));
    noconfoundersConstraint = ((argc > 4) ? ((int) atoi(argv[4])) : (0));
    
    // Read and parse the input file
    printf("# Reading the model file.\n");
    if (ParseFile(argv[1], &Btrue, &Ctrue, &Emat ))
    {
      printf("# ERROR reading the input file. Aborting.\n");
      exit(EXIT_FAILURE);
    }
    printf("# Done!\n");
    
    n = Btrue.rows;
    nexp = Emat.cols;
  }
  
  // --- Handle the case where the model is randomly generated ---
  else if ( (argc >= 6) && (argc <= 10) )
  {
    
    // number of variables
    n = (int) atoi(argv[1]);
    
    // number of experiments
    nexp = (int) atoi(argv[2]);
    
    // proportion of direct effects present
    double pdir = (double) atof(argv[3]);
    
    // propoertion of confounders present
    double pconf = (double) atof(argv[4]);
    
    // initialize random number generator
    randSeed = (int) atoi(argv[5]);
    if (randSeed == -1)
    {
      // set the random seed based on the clock (so different each time)
      randSeed = InitRand();
    }
    else
    {
      // manually set the random number seed, for reproducibility
      InitRandWithSeed( randSeed );
    }
    
    // Do we want to use R output format or not?
    int rformat = (int) atoi(argv[6]);
    SetRformat( rformat );

    // Get optional arguments
    maxcondsetsize = ((argc > 7) ? ((int) atoi(argv[7])) : (-1));
    acyclicityConstraint = ((argc > 8) ? ((int) atoi(argv[8])) : (0));
    noconfoundersConstraint = ((argc > 9) ? ((int) atoi(argv[9])) : (0));
        
    // --- create the true model ---
        
    // create a random model
    Btrue = CreateRandomB( n, pdir );
    Ctrue = CreateRandomC( n, pconf );
    
    // If acyclicity is assumed, then make sure the model generated satisfies this
    if (acyclicityConstraint)
    {
      // First make Btrue lower triangular
      for (int i=0; i<n; i++)
      {
        for (int j=i; j<n; j++)
        {
          el(Btrue, i, j) = 0;
        }
      }

      // Next create a random permutation of the variables so that 0 -> 1 -> ... -> n-1
      // is not necessarily a causal order of the variables
      matrix p = CreateMatrix(n, 1, 0);
      for (int k=0; k<n; k++) el(p, k, 0) = k;
      for (int k=0; k<n; k++)
      {
        int j = rand() % (k+1);
        int temp = el(p, j, 0);
        el(p, j, 0) = el(p, k, 0);
        el(p, k, 0) = temp;
      }
      
      matrix Btruecopy = CopyMatrix(Btrue);
      for (int i=0; i<n; i++)
      {
        for (int j=0; j<n; j++)
        {
          el(Btrue, i, j) = el(Btruecopy, el(p,i,0), el(p,j,0));
        }
      }
      
      RemoveMatrix(Btruecopy);
    }
    
    // If no confounders is assumed, make sure the model generated satisfies this
    if (noconfoundersConstraint)
    {
      for (int i=0; i<n; i++)
      {
        for (int j=0; j<n; j++)
        {
          el(Ctrue, i, j) = 0;
        }
      }
    }
    
    // create the experiments
    Emat = CreateRandomExperiments( n, nexp );
    
  }
  
  // Neither syntax was used, so state the correct syntax to the user
  else
  {
    printf("SYNTAX:\n");
    printf("discoverer filename [maxcondsetsize [acyclic [noconfounders]]]\n");
    printf("discoverer n nexp pdir pconf seed rformat [maxcondsetsize [acyclic [noconfounders]]]\n");
    exit(EXIT_FAILURE);
  }

  // Make sure maxcondsetsizeglobal has an appropriate value. -1 implies n-2.

  if (maxcondsetsize <= -1) maxcondsetsize = n-2;
  if (maxcondsetsize > n-2)
  {
    printf("# maxcondsetsize > n-2. using n-2.\n");
    maxcondsetsize = n-2;
  }
  
  // --- print the true model and the experiments, and the options ---
  
  PrintMatrix("Btrue = ", Btrue );
  PrintMatrix("Ctrue = ", Ctrue );
  PrintMatrix("Emat = ", Emat );
  printf("maxcondsetsize = %d\n",maxcondsetsize);
  
  if (acyclicityConstraint) printf("# Assuming ACYCLIC underlying model! \n");
  if (noconfoundersConstraint) printf("# Assuming NO CONFOUNDERS in the true model! \n");
  
  // --- Various initializations ---
  
  InitEdgeIndex( n, Emat );
  InitIndeps( n );
  
  // --- Initialize the independence oracle ---
  
  InitOracle(Btrue, Ctrue, Emat, maxcondsetsize);
      
  // --- Perform the actual inference ---
  
  InferModel(Emat, maxcondsetsize, acyclicityConstraint, noconfoundersConstraint);
    
  // --- output random number seed used ---
  
  printf("randSeed = %i \n", randSeed);
  
  // return 0 to indicate normal exit
  return 0;
  
}

