#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>

#include "minisatbb3api.h"

// file pointers for reading from and writing to minisatbb3
static FILE *fp;
static FILE *writefp;

// for keeping track if we are inside load or not
static int insideload = -1;

// command counter
static int cmdissued = 0;

static char longreadbuf[10000000];

// Usage:
//
// Before starting, make sure minisatbb3 is in a directory listed in the
// shell PATH variable. You can see which (if any) executable is in the path by
// typing "which minisatbb3" (without the quotes) in the shell. See the README
// file.
//
// First, start up minisatbb3 and open the connection by calling
//
// satinit( "tree" );
//
// Then, you can add constraints for example by
//
// long long vars[100];
// satload();
// vars[0] = 23;
// vars[1] = -42;
// satconstraint( 2, vars );
// vars[0] = -23;
// vars[1] = 31;
// vars[2] = 102;
// satconstraint( 3, vars );
// satloadend();
//
// This would correspond to adding the following two lines of CNF:
// 23 -42 0
// -23 31 102 0
//
// Note that:
// * You allocate a 'long long' array 'vars' to hold a constraint definition
// * A constraint definition takes the form of a list of variables either negated
//   or not, corresponding precisely to the corresponding line of CNF.
// * The 'satconstraint' function takes as input the number of elements and then
//   the pointer to the array.
// * You enclose all loading inside a beginning 'satload' and an ending 'satloadend'
//   command
//
// Then, you can issue a solve command
//
// vars[0] = 23;
// vars[1] = 24;
// vars[2] = 42;
// vars[3] = 102;
// int res = satsolve( 4, vars );
//
// The result 'res' is -1 on error, 0 on UNSAT, and 1 on SAT. If it is SAT, then
// the elements of vars contain the satisfying assignment (0's and 1's) corresponding
// to the variables. For silly implementational reasons it is required that the
// variable numbers specified are monotonically increasing (sorry!).
//
// Similarly, you can issue a backbone command
//
// vars[0] = 23;
// vars[1] = 24;
// vars[2] = 42;
// vars[3] = 102;
// int res = satbackbone( 4, vars );
//
// again returning -1, 0, and 1 (as for 'satsolve'). In this case, however, if it
// is SAT, then the elements of vars are 0 for 'always off', 1 for 'always on', and
// -1 for undetermined. For silly implementational reasons it is required that the
// variable numbers specified are monotonically increasing (sorry!).
//
// Finally, you quit minisatbb3 by calling
//
// satquit();
//
// Note that incremental solving is fully supported, so you are free to add some
// constraints, do a solve or backbone, then add some more constraints, do another
// solve or backbone, and so on. Just remember to enclose all the adding of constraints
// inside a satload(); and satloadend(); segment, and only call satsolve or satbackbone
// *outside* such segments.
//
// Please see the README file in the minisatbb3 directory for more information on
// minisatbb3.

//------------------------- public functions ----------------------

// starts up and initializes the minisatbb3 process
void satinit( const char *reordertype )
{
  // setup the pipes
  int p2c[2],c2p[2];
  if ((pipe(p2c) == -1) || (pipe(c2p) == -1))
  {
    fprintf(stderr,"Error creating the pipes. Exiting!\n");
    exit(1);
  }

  // helps remember which is which end on a pipe
  static const int readend = 0;
  static const int writeend = 1;
  
  // fork and execute minisatbb3
  pid_t childpid;
  childpid = fork();
  if (childpid >= 0)
  {
    if (childpid == 0) {
      
      // we are the child
      
      // close the pipe ends we do not use here
      close(c2p[readend]);
      close(p2c[writeend]);
      
      // make stdin the same as p2c[readend]
      close(0);
      dup(p2c[readend]);
      close(p2c[readend]);
      
      // make stdout the same as c2p[writeend]
      close(1);
      dup(c2p[writeend]);
      close(c2p[writeend]);
      
      // start minisatbb3
      execlp("minisatbb3", "minisatbb3", (char *) NULL);
      
      // if we get here, there is a problem
      fprintf(stderr,"Launching minisatbb3 failed! Exiting.\n");
      exit(1);
      
    }
    else
    {
      
      // we are the parent
      
      // close the pipe ends we do not use here
      close(c2p[writeend]);
      close(p2c[readend]);
      
      // get a file pointer to use for reading with fgets()
      fp = fdopen(c2p[readend], "r");
      if (!fp)
      {
        fprintf(stderr,"Could not open pipe for reading! Exiting.\n");
        exit(1);
      }
      
      // get a file pointer to use for writing with fwrite()
      writefp = fdopen(p2c[writeend], "w");
      if (!writefp)
      {
        fprintf(stderr,"Could not open pipe for writing! Exiting.\n");
        exit(1);
      }
        
      // note that we are currently not inside a LOAD
      insideload = 0;
      
      // check that reordertype is valid
      if (strcmp(reordertype,"none") &&
          strcmp(reordertype,"array") &&
          strcmp(reordertype,"tree"))
      {
        fprintf(stderr,"REORDER type invalid! Exiting.\n");
        exit(1);        
      }
          
      // write the REORDER command
      char cmd[100];
      sprintf(cmd, "REORDER %s", reordertype);
      satwritecommand(cmd);
      
      // wait for the corresponding OK response
      if (!satwaitresponse( 1, 10 ))
      {
        fprintf(stderr,"Not appropriate response to REORDER! Exiting.\n");
        exit(1);        
      }
            
      // return to calling function
      return;
    }
  }
  else
  {
    fprintf(stderr,"Fork failed! Exiting.\n");
    exit(1);
  } 
}


// begins a LOAD operation
void satload( void )
{
  if (insideload>0)
  {
    fprintf(stderr,"SAT load already in progress!? Exiting.\n");
    exit(1);
  }
  
  // start a load and note the command number
  insideload = satwritecommand("LOAD");

  return;
}

// adds a constraint
void satconstraint( const int n, const long long * vars )
{
  // check that we are inside a load
  if (insideload<1)
  {
    fprintf(stderr,"SAT load NOT in progress!? Exiting.\n");
    exit(1);
  }

  // this will hold the length of the line
  int linelength = 0;
  
  // this will hold an individual variable number
  char varstr[100];
  
  for (int i=0; i<n; i++)
  {
    sprintf(varstr, "%lld ", *vars);
    linelength += strlen( varstr );
    if (linelength >= (10000-10))
    {
      fprintf(stderr,"Too long line for minisatbb3! Exiting.\n");
      exit(1);      
    }
    
    fprintf(writefp, "%s", varstr);
    vars++;
  }
  fprintf(writefp, "0\n");
      
  return;
}

// ends a LOAD operation
void satloadend( void )
{
  // check to make sure we are in a load
  if (insideload<1)
  {
    fprintf(stderr,"SAT load NOT in progress!? Exiting.\n");
    exit(1);    
  }
  
  // write out the end
  fprintf(writefp, "c END\n");
  fflush(writefp);
  
  // wait for the response
  if (!satwaitresponse( insideload, 0 ))
  {
    fprintf(stderr,"Not appropriate response to LOAD! Exiting.\n");
    exit(1);
  }

  // note that we are not anymore inside a load
  insideload = 0;
    
  return;
}

// performs a SOLVE operation
int satsolve( const int n, long long * vars )
{
  return satsolve_or_backbone( n, vars, 0 );
}

// performs a BACKBONE operation
int satbackbone( const int n, long long * vars )
{
  return satsolve_or_backbone( n, vars, 1 );
}

// instructs the minisatbb3 process to QUIT
void satquit( void )
{
  int cmdnum = satwritecommand("QUIT");
  if (!satwaitresponse( cmdnum, 10 ))
  {
    fprintf(stderr,"Not appropriate response to QUIT! Exiting.\n");
    exit(1);
  }
    
  return;
}

// converts from long long array to character string
int satarray2string( int n, const long long *vars, int maxlen, char *str )
{
  // initialize
  char *tempstring = (char *) malloc(255);
  strcpy(str, "");
  long long startrange = 0;
  int nowlen = 0;

  for (int i=0; i<n; i++)
  {
    if (i==0)
    {
      // we just need to get started
      startrange = vars[0];
    }
    else if (startrange && (vars[i] == vars[i-1]+1))
    {
      // we are in a range and have consecutive variables so do nothing
      continue;
    }
    else if (startrange && (vars[i] != vars[i-1]+1))
    {
      // we are in a range but do not have consecutive variables so
      // we need to cut the range and start a new one
    
      if (startrange == vars[i-1])
      {
        // only one element
        sprintf(tempstring,"%lld,",startrange);
      }
      else
      {
        // range
        sprintf(tempstring,"%lld-%lld,",startrange,vars[i-1]);
      }
      
      if (nowlen + strlen(tempstring) >= (maxlen-1))
      {
        free(tempstring);
        return 0;
      }
      else
      {
        strcat(str,tempstring);
        nowlen += strlen(tempstring);
        startrange = vars[i];
      }
      
    }
  }
  
  if (n>0)
  {
    // finally we need to end the string
    if (startrange == vars[n-1])
    {
      // only one element
      sprintf(tempstring,"%lld",startrange);
    }
    else
    {
      // range
      sprintf(tempstring,"%lld-%lld",startrange,vars[n-1]);
    }
    
    if (nowlen + strlen(tempstring) >= (maxlen-1))
    {
      free(tempstring);
      return 0;
    }
    else
    {
      strcat(str,tempstring);
      nowlen += strlen(tempstring);
    }
  }
    
  free(tempstring);
  return 1;
  
}

// converts from character string to long long array
// pass the maximum number of elements that fits in the array in *n.
// the number actually put in the array is put there upon successful return
// the function returns 1 on success and 0 on error
int satstring2array( int *n, long long *vars, const char *str )
{
  char thisrange[1000];
  char *cptr;
  
  // maximum number of elements that can fit in the array
  int maxn = *n;
  
  // number of elements read so far
  int nelem = 0;
    
  while (strlen(str)>0)
  {
    // Get part of string up to and including the first comma (',') character or
    // until the end if no comma is encountered. Leave the rest.
    if ((cptr = strchr(str, ',')))
    {
      strncpy(thisrange,str,(cptr-str));
      thisrange[(cptr-str)] = '\0';
      str = cptr+1;
    }
    else
    {
      strcpy(thisrange,str);
      str += strlen(str);
    }
        
    // Get the range
    long long rangeMin;
    long long rangeMax;
    if (strchr(thisrange,'-'))
    {
      sscanf(thisrange, "%lld-%lld", &rangeMin, &rangeMax);
    }
    else
    {
      sscanf(thisrange, "%lld", &rangeMin );
      rangeMax = rangeMin;
    }
        
    for (long long externVar = rangeMin; externVar <= rangeMax; externVar++)
    {
      if (nelem >= maxn)
      {
        fprintf(stderr,"Could not fit range into array.");
        *n = nelem;
        return 0;
      }
      
      vars[nelem] = externVar;
      nelem++;
    }
  }
  
  *n = nelem;
  return 1;
}


//------------------------- internal functions ----------------------

int readresponseline( char *readbuf, int maxlen, FILE *source, int waitsecs )
{
  int nsec = 0;
  while (fgets(readbuf, maxlen, source) == NULL)
  {
    sleep(1);
    nsec++;
    if ((waitsecs>0) && (nsec > waitsecs))
    {
      fprintf(stderr,"Waited in vain %d seconds for response from minisatbb3.\n", waitsecs);
      return 1;
    }
  }
  return 0; 
}


// writes a command
int satwritecommand( const char * str )
{
  // get new command number and write out the command
  cmdissued++;
  fprintf(writefp, "%d %s\n", cmdissued, str);
  fflush(writefp);
  
  // return the command number
  return cmdissued;
}

// waits for a OK response, set waitsecs=0 to wait forever
int satwaitresponse( int cmdnum, int waitsecs )
{
  char readbuf[100];
  if (readresponseline(readbuf, sizeof(readbuf), fp, waitsecs)) return 0;
  char responsestr[100];
  sprintf(responsestr, "%d OK\n", cmdnum);
  if (strcmp(responsestr, readbuf))
  {
    fprintf(stderr,"Response from minisatbb3 was not appropriate.\n");
    fprintf(stderr,"It was: %s",readbuf);
    return 0;
  }
  
  return 1;
}

// internal function for handling both SOLVE and BACKBONE
int satsolve_or_backbone( const int n, long long * vars, const int bbflag )
{
  // create the initial command string and issue the command
  int cmdnum;
  if (bbflag) cmdnum = satwritecommand( "BACKBONE" );
  else cmdnum = satwritecommand( "SOLVE" );
  
  // create the list of variables
  char *varstr;
  int maxlength = 255;
  while (1)
  {
    varstr = malloc(maxlength);
    if (satarray2string( n, vars, maxlength, varstr ))
    {
      // success
      break;
    }
    else
    {
      // failure (string not long enough)
      free(varstr);
      maxlength = maxlength*10;
    }
  }
  
  // write the list of variables and free the memory used for the list of variables
  fprintf(writefp, "%s\n", varstr);
  fflush(writefp);
  free(varstr);
      
  // read the first line of the response (will be SAT, UNSAT, or INDET)
  char readbuf[100];
  if (readresponseline(readbuf, sizeof(readbuf), fp, 0)) return -1;
  if (!strcmp("INDET\n", readbuf))
  {
    // say that it was indet
    fprintf(stderr,"Received INDET response from minisatbb3. Returning -1.\n");

    // next line will be the "# OK" response, need to read it as well
    satwaitresponse( cmdnum, 0 );

    return -1;
  }
  else if (!strcmp("UNSAT\n", readbuf))
  {
    // say that it was unsat
    fprintf(stderr,"It was UNSAT.\n");
    
    // next line will be the "# OK" response, need to read it as well
    satwaitresponse( cmdnum, 0 );
    
    // return 0 to indicate we got "UNSAT"
    return 0;
  }
  else if (!strcmp("SAT\n", readbuf))
  {
    // read and print the next line
    char *cptr = longreadbuf;
    int nbytes;
    long long x;
    if (readresponseline(longreadbuf, sizeof(longreadbuf), fp, 0)) return -1;

    // go through each variable for which a response was requested
    for (int i=0; i<n; i++) 
    {
      // get the variable number
      long long varnum = vars[i];
      
      // read in the current numbered variable (and check for a plus sign before the number)
      int ispositive = 0;
      if (sscanf(cptr, "+%lld %n", &x,&nbytes)) ispositive = 1;
      else sscanf(cptr, "%lld %n", &x,&nbytes); 
              
      // was the (absolute value) of the number we read different from the array element?
      if (abs(x) != varnum)
      {
        // yes! just mark this number as OFF (for SOLVE) or nonbackbone (for BACKBONE)
        if (bbflag) vars[i] = -1;
        else vars[i] = 0;
        
        // go to the next number (without incrementing cptr)
        continue;       
      }
      else 
      {
        // number read equals the one in the array, so mark it appropriately
        if (bbflag)
        {
          if (ispositive) vars[i] = 1;
          else if (x<0) vars[i] = 0;
          else vars[i] = -1;
        }
        else 
        {
          if (x<0) vars[i] = 0;
          else vars[i] = 1;
        }
        
        // increment cptr
        cptr += nbytes;
        
      }
    }
    
    // next line will be the "# OK" response, need to read it as well
    satwaitresponse( cmdnum, 0 );
    
    // return 1 to indicate we got "SAT"
    return 1;
  }
  else 
  {
    // none of the above, this is an error!
    fprintf(stderr,"Got response: %s",readbuf);
    fprintf(stderr,"BACKBONE or SOLVE returned none of the valid responses! Exiting.");
    exit(1);
  }
  
  return 0;
}




