
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "matrix.h"
#include "indeps.h"
#include "oracle.h"
#include "edgepathindex.h"

// We store the dseps here
static matrix trueModelDseps;

void InitOracle(matrix Btrue, matrix Ctrue, matrix E, int maxcondsetsize)
{
  // Create the matrix to store the dseps
  int nexp = E.cols;
  long long nindep = nTotIndep();
  trueModelDseps = CreateMatrix(nindep, nexp, 3);
  
  // For simplicity, we here simply compute *all possible* dseps
  // (this is feasible up to around n=15 variables or so)
  ComputeDSeps( Btrue, Ctrue, E, trueModelDseps, 0, maxcondsetsize );
    
  return;
}

// This takes as input the problem parameters (B, C, E), and a result-matrix
// (of size nindep-by-nexp). The dsep info is placed here, with 0=d-connected,
// 1=d-separated, 3=unobservable or out of requested conditioning set size.
// The final arguments state the minimum and maximum conditioning set sizes
// to compute (and they can be identical to compute just a single one).

void ComputeDSeps( matrix Btrue, matrix Ctrue, matrix E, matrix resDsep,
                   int mincsetsize, int maxcsetsize )
{
  // Get the number of variables and the number of experiments
  int n = E.rows;
  int nexp = E.cols;

  // Initialize auxiliary matrices to use here
  matrix colliderok = CreateMatrix( 1, n, 0 );
  matrix sub = CreateMatrix(1, n, 0);
  matrix ivec = CreateMatrix(1, n, 0);
  matrix Amat = CreateMatrix( n, n, 0);
  matrix Cmat = CreateMatrix( n, n, 0);
  matrix Dmat = CreateMatrix( n, n, 0);
  
  // Do one experiment at a time
  for (int thisexp=0; thisexp<nexp; thisexp++)
  {
    
    // Construct the manipulated model
    matrix Bmani = CopyMatrix( Btrue );
    matrix Cmani = CopyMatrix( Ctrue );
    for (int i=0; i<n; i++)
    {
      // Is variable i intervened on?
      if (el(E,i,thisexp)==2)
      {
        // Remove any direct edges *into* i
        for (int j=0; j<n; j++) el(Bmani, i, j) = 0;
        
        // Remove any confounding edges between i and any other variable
        for (int j=0; j<n; j++)
        {
          el(Cmani, i, j) = 0;
          el(Cmani, j, i) = 0;
        }
      }
    }
    
    // Do one conditioning set at a time
    int nSubsets = nTotSub();
    for (int si = 0; si < nSubsets; si++)
    {
      SubIndexToSubVec( si, sub );
      
      // Get the number of elements in the subset
      int ss = SubSize(sub);
      
      // If the subset is of size larger than n-2 then go to next one
      if (ss > (n-2)) continue;
      
      // If the subset is smaller than mincsetsize or larger than maxcsetsize
      // then just go to the next one
      if ((ss < mincsetsize) || (ss > maxcsetsize)) continue;
      
      // Which nodes are in the conditioning set or ancestors of those
      // in the conditioning set?
      SetMatrix( colliderok, 0 );
      for (int i=0; i<n; i++) if (el(sub,0,i)) el(colliderok,0,i)=1;
      int changed = 1;
      while (changed)
      {
        changed = 0;
        for (int i=0; i<n; i++)
        {
          if (el(colliderok,0,i)==0)
          {
            for (int j=0; j<n; j++)
            {
              if (j==i) continue;
              if ((el(Bmani,j,i)==1) && (el(colliderok,0,j)))
              {
                el(colliderok,0,i) = 1;
                changed = 1;
                break;
              }
            }
          }
        }
      }
      
      // Construct the matrices to hold the existence of different kinds
      // of d-connecting paths
      // A: x = ... = y, C: x = ... => y, D: x <= ... => y.
      // with C[x,y] being a path with a tail at x and an arrowhead at y.
      // (Note that each path is represented twice, corresponding to permuting x and y,
      // in A and D, but let's not worry about this little bit of redundancy.)
      SetMatrix(Amat, 0);
      SetMatrix(Cmat, 0);
      SetMatrix(Dmat, 0);
      
      // Go through all direct edges
      for (int startNode = 0; startNode < n; startNode++)
      {
        for (int stopNode = 0; stopNode < n; stopNode++)
        {
          // If startNode = stopNode go to next one
          if (startNode == stopNode) continue;
          
          // If it does not exist in the model, skip it
          if (el(Bmani,stopNode,startNode) == 0) continue;
          
          // Add it to the matrices B and C
          el(Cmat,startNode,stopNode) = 1;
        }
      }
      
      // Go through all confounders
      for (int startNode = 0; startNode < n; startNode++)
      {
        for (int stopNode = startNode+1; stopNode < n; stopNode++)
        {
          // If it does not exist in the model, skip it
          if (el(Cmani,stopNode,startNode) == 0) continue;
          
          // Add it to the matrix D
          el(Dmat,startNode,stopNode) = 1;
          el(Dmat,stopNode,startNode) = 1;
        }
      }
            
      // Extend paths to find all d-connecting paths
      {
        int changed = 1;
        while (changed)
        {
          changed = 0;
          
          // Try to add entries to Amat
          for (int i=0; i<n; i++)
          {
            for (int j=0; j<n; j++)
            {
              if (j==i) continue;
              if (el(Amat,i,j)==0)
              {
                for (int k=0; k<n; k++)
                {
                  if ((k==i) || (k==j)) continue;
                  if ((el(Amat,i,k) && el(Amat,k,j) && (el(sub,0,k)==0)) ||
                      (el(Cmat,i,k) && el(Amat,k,j) && (el(sub,0,k)==0)) ||
                      (el(Cmat,i,k) && el(Cmat,j,k) && el(colliderok,0,k)))
                  {
                    el(Amat,i,j) = 1;
                    el(Amat,j,i) = 1;
                    changed = 1;
                    break;
                  }
                }
              }
            }
          }
          
          // Try to add entries to Cmat
          for (int i=0; i<n; i++)
          {
            for (int j=0; j<n; j++)
            {
              if (j==i) continue;
              if (el(Cmat,i,j)==0)
              {
                for (int k=0; k<n; k++)
                {
                  if ((k==i) || (k==j)) continue;
                  if ((el(Amat,i,k) && el(Cmat,k,j) && (el(sub,0,k)==0)) ||
                      (el(Cmat,i,k) && el(Cmat,k,j) && (el(sub,0,k)==0)) ||
                      (el(Cmat,i,k) && el(Dmat,k,j) && (el(colliderok,0,k))) ||
                      (el(Amat,i,k) && el(Dmat,k,j) && (el(sub,0,k)==0)))
                  {
                    el(Cmat,i,j) = 1;
                    changed = 1;
                    break;
                  }
                }
              }
            }
          }
          
          // Try to add entries to Dmat
          for (int i=0; i<n; i++)
          {
            for (int j=0; j<n; j++)
            {
              if (j==i) continue;
              if (el(Dmat,i,j)==0)
              {
                for (int k=0; k<n; k++)
                {
                  if ((k==i) || (k==j)) continue;
                  if ((el(Cmat,k,i) && el(Cmat,k,j) && (el(sub,0,k)==0)) ||
                      (el(Dmat,i,k) && el(Cmat,k,j) && (el(sub,0,k)==0)) ||
                      (el(Cmat,k,i) && el(Dmat,k,j) && (el(sub,0,k)==0)) ||
                      (el(Dmat,i,k) && el(Dmat,k,j) && el(colliderok,0,k)))
                  {
                    el(Dmat,i,j) = 1;
                    el(Dmat,j,i) = 1;
                    changed = 1;
                    break;
                  }
                }
              }
            }
          }
        }
        
        // Harvest the found dsep information

        for (int startNode=0; startNode<n; startNode++)
        {
          for (int stopNode=startNode+1; stopNode<n; stopNode++)
          {
            // Get the indep vec
            SetMatrix( ivec, 0 );
            for (int i=0; i<n; i++)
            {
              if ((i == startNode) || (i == stopNode)) el(ivec,0,i) = 2;
              else if (el(sub, 0, i)==1) el(ivec,0,i) = 1;
            }
            
            // Get the corresponding index
            int ii = IndepVecToIndepIndex( ivec );
            
            // If either of the endpoints or any in the conditioning set
            // are hidden in this experiment, then simply store a '3'
            int hidden=0;
            for (int i=0; i<n; i++)
            {
              if ((el(E,i,thisexp)==0) && (el(ivec, 0, i)>0))
              {
                hidden=1;
              }
            }
            if (hidden)
            {
              el(resDsep,ii,thisexp) = 3;
            }
            else
            {
              // Not hidden, so store the dsep: If any d-connecting path has been
              // found, mark as dependent, else independent.
              if (el(Amat,startNode,stopNode) || el(Cmat,startNode,stopNode) ||
                  el(Cmat,stopNode,startNode) || el(Dmat,startNode,stopNode))
              {
                el(resDsep,ii,thisexp) = 0;
              }
              else
              {
                el(resDsep,ii,thisexp) = 1;
              }
            }
          }
        }
      }
      
    }
  }

  // Deallocate the space for these matrices
  RemoveMatrix(Amat);
  RemoveMatrix(Cmat);
  RemoveMatrix(Dmat);
  RemoveMatrix(colliderok);
  RemoveMatrix(sub);
  RemoveMatrix(ivec);
  
  return;
}

int DSepOracleIndex( long long ii, long long ei )
{
  return el(trueModelDseps, ii, ei);
}

