#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "matrix.h"
#include "utils.h"

// --- Various utility functions ---

// Initialize random number generator
int InitRand( void )
{
  int randSeed = time(NULL);
  srand( randSeed );
  return randSeed;
}

void InitRandWithSeed( int randSeed )
{
  srand( randSeed );
  return;
}

// --- Functions to create a random true model ---

// Creates and returns an n-by-n binary matrix with a zero diagonal,
// where element (i,j) is 1 if there is an edge from variable j to variable i,
// and 0 if there is no such edge.
matrix CreateRandomB( int n, double pdir )
{
  matrix B;
  int i,j;
  
  // Create the matrix
  B = CreateMatrix( n, n, 0 );
  
  // Randomly assign each edge to exist (1) or not (0)
  for (i=0; i<n; i++)
  {
    for (j=0; j<n; j++)
    {
      el(B,i,j) = ((rand() % 100)<((int) (100.0 * pdir)))?(1):(0);
    }
  }
  
  // Make sure the diagonal is zero
  for (i=0; i<n; i++)
  {
    el(B,i,i) = 0;
  }
  
  return B;
}

// Creates and returns an n-by-n symmetric binary matrix with a zero diagonal,
// where element (i,j) is 1 if there is a confounder between variable j and variable i,
// and 0 if there is no such confounder.
matrix CreateRandomC( int n, double pconf )
{
  matrix C;
  int i,j;
  
  // Create the matrix
  C = CreateMatrix( n, n, 0 );
  
  // Randomly assign each edge to exist (1) or not (0)
  for (i=0; i<n; i++)
  {
    for (j=i+1; j<n; j++)
    {
      el(C,i,j) = ((rand() % 100)<((int) (100.0 * pconf)))?(1):(0);
      el(C,j,i) = el(C,i,j);
    }
  }
  
  // Make sure the diagonal is zero
  for (i=0; i<n; i++)
  {
    el(C,i,i) = 0;
  }
  
  return C;
}

// --- Function to create a random experiments matrix ---

// Creates and returns a matrix with n rows and nexp columns, where
// each column denotes one experiment. A value of 2 indicates that the
// variable in question is intervened on, a value of 1 indicates observation,
// and a value of 0 indicates that the variable is hidden (latent).
matrix CreateRandomExperiments( int n, int nexp )
{
  matrix E;
  int i,j,nones,ntwos;
  
  // Create the matrix
  E = CreateMatrix( n, nexp, 0 );
  
  // Do one experiment at a time
  for (i=0; i<nexp; i++)
  {
    while (1) {
      
      // Create the experiment (0 = unobserved, 1 = observed, 2 = intervened),
      // keeping track of the total number of 1's and 2's
      nones = 0;
      ntwos = 0;
      for (j=0; j<n; j++)
      {
        el(E,j,i) = (rand() % 3);
        if (el(E,j,i) == 1) nones++;
        if (el(E,j,i) == 2) ntwos++;
      }
      
      // If we have at least two non-unobserved variables, and at least one
      // observed variable, this is a reasonable experiment and we can exit
      if (((nones+ntwos) >= 2) && (nones >= 1)) break;
      
    }
  }
  
  return E;
}

// --- Various helper functions ---

// reads in and parses the input (R-format) file
int ParseFile(const char *fname, matrix *B, matrix *C, matrix *E)
{
  int n = 0, nexp = 0;
  matrix Btrue, Ctrue, Emat;
  
  // Open the file
  FILE *thefile = fopen( fname, "r" );
  
  // If we could not open it, return an error
  if (!thefile)
  {
    printf("# Could not open the file.\n");
    return 1;
  }
  
  // Parse the file
  {
    char line[1000];
    while(fgets(line, 1000, thefile) != NULL)
    {
      // If the line starts with '#' then ignore it
      if (line[0] == '#') continue;
      
      // Check for the definition of Btrue
      else if (!strncmp(line, "Btrue =", 7))
      {
        int dummy;
        
        // Get the dimensions
        fgets(line, 1000, thefile);
        sscanf(line, "%i %i", &n, &dummy);
        if (dummy != n)
        {
          printf("# Btrue is not square!\n");
          return 1;
        }
        
        // Create the matrix
        Btrue = CreateMatrix( n, n, 0 );
        
        // Get the file data
        for (int i=0; i<n; i++)
        {
          for (int j=0; j<n; j++)
          {
            // Read the one or the zero
            int c = fgetc(thefile);
            if (((char) c) == '1')
            {
              el(Btrue,i,j) = 1;
            }
            // Handle the space or newline
            (void) fgetc(thefile);
          }
        }
      }
      
      // Check for the definition of Ctrue
      else if (!strncmp(line, "Ctrue =", 7))
      {
        int dummy;
        
        // Get the dimensions
        fgets(line, 1000, thefile);
        sscanf(line, "%i %i", &n, &dummy);
        if (dummy != n)
        {
          printf("# Ctrue is not square!\n");
          return 1;
        }
        
        // Create the matrix
        Ctrue = CreateMatrix( n, n, 0 );
        
        // Get the file data
        for (int i=0; i<n; i++)
        {
          for (int j=0; j<n; j++)
          {
            // Read the one or the zero
            int c = fgetc(thefile);
            if (((char) c) == '1')
            {
              el(Ctrue,i,j) = 1;
            }
            // Handle the space or newline
            (void) fgetc(thefile);
          }
        }
      }
      
      // Check for the definition of E
      else if (!strncmp(line, "E =", 3))
      {
        // Get the dimensions
        fgets(line, 1000, thefile);
        sscanf(line, "%i %i", &n, &nexp);
        
        // Create the matrix
        Emat = CreateMatrix( n, nexp, 0 );
        
        // Get the file data
        for (int i=0; i<n; i++)
        {
          for (int j=0; j<nexp; j++)
          {
            // Read the one or the zero
            int c = fgetc(thefile);
            if (((char) c) == '1')
            {
              el(Emat,i,j) = 1;
            }
            else if (((char) c) == '2')
            {
              el(Emat,i,j) = 2;
            }
            // Handle the space or newline
            (void) fgetc(thefile);
          }
        }
      }
      
      else
      {
        continue;
      }
      
    }
  }
  fclose(thefile);
    
  // Store the read matrices as requested
  *B = Btrue;
  *C = Ctrue;
  *E = Emat;
  
  // Return 0 to indicate everything ok
  return 0;
}

int zInBetween( matrix Btest, matrix Ctest, int x, int y, int z )
{
  int n = Btest.rows;
  int connectedDirectlyToX;
  int connectedDirectlyToY;
  
  {
    // Construct connection matrix
    matrix Bconn = CreateMatrix(n, n, 0);
    for (int i=0; i<n; i++)
    {
      for (int j=0; j<n; j++)
      {
        if ((el(Btest,i,j)>0) || (el(Btest,j,i)>0) || (el(Ctest,i,j)>0))
        {
          el(Bconn,i,j) = 1;
        }
      }
    }
    
    // Take out any connections to and from y
    for (int i=0; i<n; i++)
    {
      el(Bconn,i,y) = 0;
      el(Bconn,y,i) = 0;
    }
    
    // Check direct connectedness
    connectedDirectlyToX = connectedNodes( Bconn, x, z );
  }
  
  {
    // Construct connection matrix
    matrix Bconn = CreateMatrix(n, n, 0);
    for (int i=0; i<n; i++)
    {
      for (int j=0; j<n; j++)
      {
        if ((el(Btest,i,j)>0) || (el(Btest,j,i)>0) || (el(Ctest,i,j)>0))
        {
          el(Bconn,i,j) = 1;
        }
      }
    }
    
    // Take out any connections to and from x
    for (int i=0; i<n; i++)
    {
      el(Bconn,i,x) = 0;
      el(Bconn,x,i) = 0;
    }
    
    // Check direct connectedness
    connectedDirectlyToY = connectedNodes( Bconn, y, z );
  }
  
  if (connectedDirectlyToX && connectedDirectlyToY) return 1;
  else return 0;
}

int connectedNodes( matrix M, int x, int y )
{
  // Input:
  // - A binary symmetric n-by-n matrix M, where 1=edge, 0=no edge.
  // - A node x, with 0 <= x <= n-1
  // - A node y, with 0 <= y <= n-1
  // Outputs:
  // 1 if nodes x and y are connected in the graph
  // 0 if they are disconnected.
  // Algorithm:
  // Starts at x and marks reachable nodes one by one. Not optimized,
  // this will not be a critical component.
  
  int n = M.rows;
  matrix R = CreateMatrix( 1, n, 0);
  int progressmade = 1;
  el(R, 0, x) = 1;
  while (progressmade)
  {
    progressmade = 0;
    for (int i=0; i<n; i++)
    {
      if (el(R, 0, i) == 0)
      {
        for (int j=0; j<n; j++)
        {
          if ((el(R, 0, j) == 1) && (el(M, i, j) == 1))
          {
            el(R, 0, i) = 1;
            progressmade = 1;
          }
        }
      }
    }
  }
  
  if (el(R, 0, y) == 1) return 1;
  else return 0;
}


