/****************************************************************************************[Dimacs.h]
Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson
Copyright (c) 2007-2010, Niklas Sorensson
Modifications 2012-2013 by Patrik Hoyer

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#ifndef Minisat_Dimacs_h
#define Minisat_Dimacs_h

#include <stdio.h>

#include "utils/ParseUtils.h"
#include "core/SolverTypes.h"
#include "core/Helpers.h"

namespace Minisat {


//=================================================================================================
// DIMACS Parser:

template<class Solver>
static void parse_DIMACS_main(Solver& S) {
    vec<Lit> lits;
        
    // read one line at a time
    int cnt = 0;
    for (;;){

      // initialization
      lits.clear();
      
      // read in characters one at a time
      int comment = 0;
      int willend = 1;
      char endstring[10];
      strcpy(endstring,"c END");
      bool neg = false;
      long long val = 0;
      int i;
      for (i=0;;i++) 
      {      
        char c = fgetc(stdin);
        if (c=='\n') 
        {
          // found newline so finish the number and we are ready with this line
          // note: last number is by definition a zero so no need to do anything with it
          break;
        }
        else if (comment || ((i==0) && ((c=='c') || (c=='p'))))
        {
          // comment or problem statement
          comment = 1;
          if ((i<=4) && (c != endstring[i])) willend = 0;
        }
        else if (c==' ')
        {
          // regular space so finish this number and start a new number
          if (val == 0) continue;
          int var = getInternVar( val );
          if (var == 0) var = createInternVar( val );
          var--;
          while (var >= S.nVars()) S.newVar();
          lits.push( neg ? ~mkLit(var) : mkLit(var) );
          neg = false;
          val = 0;
          willend = 0;
        }
        else if (c=='-')
        {
          // negation sign
          neg = true; 
          willend = 0;         
        }
        else if (c=='+')
        {
          // plus sign, do nothing
          willend = 0;
        }
        else if ((c >= '0') && (c <= '9'))
        {
          // digit so continue an existing number
          val = val*10 + (c - '0');
          willend = 0;
        } 
        else
        {
          fprintf(stderr, "PARSE ERROR! Unexpected char: %c\n", c);
          exit(3);
        }     
      }
      
      // check if we got the end message
      if ((i==5) && (willend))
      {
        break;
      }
    
      // increment count and add the clause
      cnt++;
      S.addClause_(lits);    
    }
    
}

template<class Solver>
static void parse_DIMACS(Solver& S) {
    parse_DIMACS_main(S); }

//=================================================================================================
}

#endif
