/****************************************************************************************[Helpers.cc]
Copyright (c) 2012-2013 by Patrik Hoyer

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "core/Helpers.h"

// This will hold the reordering type
int reordertype = 0;

// Keep track of largest used internal variable in case REORDER is none
int largestIntVar = 0;

// These will hold the permutations for reorder type 'array'
int *externToIntern;
int maxIntern;
long long limitExtern;

// These will be used for the prefix tree
static int maxused = 0;
static tnode *rootnode = NULL;

// Writes a command response to stdout
void writeResponse(int cmdNum, char *response)
{
  printf("%d %s\n", cmdNum, response);
  fflush(NULL);
  return; 
}

// Gets the next command from the command file
int getNextCmd(char *cmd)
{   
  // Try to read from stdinfp until we get a full line
  while(fgets(cmd, MAXCMDLEN, stdin) == NULL) {}
  
  // Get the command number at the beginning of the line and return it
  int cmdnumber=-1;
  sscanf(cmd, "%d", &cmdnumber);
  return cmdnumber;         
}

void setReorderType(int rtype)
{
  // if reordering set to 'none' then this setting simply needs to be stored
  if (rtype == 0)
  {
    reordertype = 0;
  }
  // if reordering set to 'array' then we use the array-based reordering method
  else if (rtype == 1)
  {
    // store this option and initialize the permutation array 
    reordertype = 1;
    initPerm();    
  }
  // if reordering set to 'tree' then we use the tree-based reordering method
  else if (rtype == 2)
  {
    // store this option 
    reordertype = 2;
  }
}

// Sets up the initial permutation array
void initPerm()
{
  maxIntern = 0;
  limitExtern = 1000;
  externToIntern = (int *) malloc( sizeof(int) * (limitExtern+1) );
  if (!externToIntern) fprintf(stderr,"MEMORY ALLOCATION ERROR in initPerm()!\n");
  for (int i=0; i<limitExtern+1; i++) externToIntern[i] = 0;
}

// Checks if a given external variable has an internal representation already
// If it does, it simply returns the internal variable number. If it does not, 
// it returns 0.
int getInternVar( long long externVar )
{
  int internVar;

  // Check to make sure that the external variable is larger than 0
  if (externVar <= 0)
  {
    fprintf(stderr,"in getInternVar: externVar was <= 0! It was: %lld \n",externVar);
    fprintf(stderr,"Exiting!\n");
    exit(0);
  }

  // If reordertype is zero then no reordering
  if (reordertype == 0)
  {
    if (externVar > largestIntVar) return 0; else return (int) externVar;
  }
  
  // If reordertype is 2 then we use tree-based approach
  if (reordertype == 2)
  {
    return trie_lookup_or_add( externVar, 0 );
  }
  
  // Is the external variable within the range?
  if (externVar <= limitExtern)
  {
    // If the variable exists, it is larger than 0, in this case we return it.
    // If we get a 0, the variable has not yet been encountered, so we return 0.
    internVar = externToIntern[externVar];
    return internVar;   
  }
  else
  {
    return 0;
  }
}

// Creates a new internal variable to hold the specified external one
int createInternVar( long long externVar )
{
  int internVar;

  // Check to make sure that the external variable is larger than 0
  if (externVar <= 0)
  {
    fprintf(stderr,"in createInternVar: externVar was <= 0! It was: %lld \n",externVar);
    fprintf(stderr,"Exiting!\n");
    exit(0);
  }

  // If reordertype is zero then no reordering
  if (reordertype == 0)
  {
    if (externVar > largestIntVar) largestIntVar = externVar;
    return (int) externVar;
  }
  
  // If reordertype is 2 then we use tree-based approach
  if (reordertype == 2)
  {
    return trie_lookup_or_add( externVar, 1 );
  }
  
  // Is the external variable within the range?
  if (externVar <= limitExtern)
  {
    // Check that it indeed did not exist yet!
    if (externToIntern[externVar])
    {
      fprintf(stderr,"externVar already existed! \n");
      fprintf(stderr,"Exiting!\n");
      exit(0);
    }
    
    // Create it
    maxIntern++;
    internVar = maxIntern;
    externToIntern[externVar] = internVar;
    return internVar;   
  }
  else
  {
    // We must extend the range! Extending it to 3 times the value of externVar.
    long long newLimitExtern = 3*externVar;
    int *newExternToIntern = (int *) malloc( sizeof(int) * (newLimitExtern+1) );
    if (!newExternToIntern) fprintf(stderr,"MEMORY ALLOCATION ERROR in createInternVar()!\n");
    for (long long i=0; i<limitExtern+1; i++) 
    {
      newExternToIntern[i] = externToIntern[i];
    }
    for (long long i=limitExtern+1; i<newLimitExtern+1; i++)
    {
      newExternToIntern[i] = 0;
    }
    free(externToIntern);
    externToIntern = newExternToIntern;
    limitExtern = newLimitExtern;
    
    // Create it
    maxIntern++;
    internVar = maxIntern;
    externToIntern[externVar] = internVar;
    return internVar;
  }
}

int trie_lookup_or_add( long long x, int addflag )
{
  // create the root node if this is the first call
  if (!rootnode)
  {
    rootnode = (tnode *) malloc( sizeof(tnode) );
    rootnode->y = 0;
    for (int i=0; i<10; i++) rootnode->ch[i] = NULL;
  }
  
  // check to make sure the integer is valid
  if (x <= 0) return -1;
  
  // convert the long integer to a string
  char str[255];
  sprintf(str, "%lld", x);
  int len = strlen(str);
  
  // step through the tree to find the relevant node
  tnode *curnode = rootnode;
  int didnotexist = 0;
  for (int i=0; i<len; i++)
  {
    // convert digit to integer
    int digit = ((int) str[i]) - ((int) '0');
    
    // does the tree extend to this?
    tnode *childnode = curnode->ch[digit];
    if (childnode)
    {
      curnode = childnode;
      continue;
    }
    else 
    {
      didnotexist = 1;
      if (addflag)
      {
        childnode = (tnode *) malloc(sizeof(tnode));
        curnode->ch[digit] = childnode;
        childnode->y = 0;
        for (int i=0; i<10; i++) childnode->ch[i] = NULL;
        curnode = childnode;
        continue;
      }
      else
      {
        break;
      }
    }
  }
    
  if (didnotexist)
  {
    if (!addflag)
    {
      return 0;
    }
    else 
    {
      maxused++;
      curnode->y = maxused;
      return maxused;
    }
  }
  else 
  {
    if (curnode->y == 0)
    {
      if (!addflag)
      {
        return 0;
      }
      else
      {
        maxused++;
        curnode->y = maxused;
        return maxused;
      }
    }
    else
    {
      return curnode->y;
    }
  }
  
}
