/*****************************************************************************************[Main.cc]
Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson
Copyright (c) 2007-2010, Niklas Sorensson
Modifications 2012-2013 by Patrik Hoyer

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <signal.h>
#include <zlib.h>

#include "utils/System.h"
#include "utils/ParseUtils.h"
#include "utils/Options.h"
#include "core/Dimacs.h"
#include "core/Solver.h"
#include "core/Helpers.h"

using namespace Minisat;

//=================================================================================================


void printStats(Solver& solver)
{
    double cpu_time = cpuTime();
    double mem_used = memUsedPeak();
    printf("restarts              : %"PRIu64"\n", solver.starts);
    printf("conflicts             : %-12"PRIu64"   (%.0f /sec)\n", solver.conflicts   , solver.conflicts   /cpu_time);
    printf("decisions             : %-12"PRIu64"   (%4.2f %% random) (%.0f /sec)\n", solver.decisions, (float)solver.rnd_decisions*100 / (float)solver.decisions, solver.decisions   /cpu_time);
    printf("propagations          : %-12"PRIu64"   (%.0f /sec)\n", solver.propagations, solver.propagations/cpu_time);
    printf("conflict literals     : %-12"PRIu64"   (%4.2f %% deleted)\n", solver.tot_literals, (solver.max_literals - solver.tot_literals)*100 / (double)solver.max_literals);
    if (mem_used != 0) printf("Memory used           : %.2f MB\n", mem_used);
    printf("CPU time              : %g s\n", cpu_time);
}


static Solver* solver;

// Terminate by notifying the solver and back out gracefully. This is mainly to have a test-case
// for this feature of the Solver as it may take longer than an immediate call to '_exit()'.
static void SIGINT_interrupt(int signum) { solver->interrupt(); }

// Note that '_exit()' rather than 'exit()' has to be used. The reason is that 'exit()' calls
// destructors and may cause deadlocks if a malloc/free function happens to be running (these
// functions are guarded by locks for multithreaded use).
static void SIGINT_exit(int signum) {
    printf("\n"); printf("*** INTERRUPTED ***\n");
    if (solver->verbosity > 0){
        printStats(*solver);
        printf("\n"); printf("*** INTERRUPTED ***\n"); }
    _exit(1); }



//=================================================================================================
// Main:


int main(int argc, char** argv)
{
    try {
        setUsageHelp("USAGE: %s [options] \n\n");
        
#if defined(__linux__)
        fpu_control_t oldcw, newcw;
        _FPU_GETCW(oldcw); newcw = (oldcw & ~_FPU_EXTENDED) | _FPU_DOUBLE; _FPU_SETCW(newcw);
        // printf("WARNING: for repeatability, setting FPU to use double precision\n");
#endif
        // Extra options:
        //
        IntOption    verb   ("MAIN", "verb",   "Verbosity level (0=silent, 1=some, 2=more).", 1, IntRange(0, 2));
        IntOption    cpu_lim("MAIN", "cpu-lim","Limit on CPU time allowed in seconds.\n", INT32_MAX, IntRange(0, INT32_MAX));
        IntOption    mem_lim("MAIN", "mem-lim","Limit on memory usage in megabytes.\n", INT32_MAX, IntRange(0, INT32_MAX));
        
        parseOptions(argc, argv, true);

        Solver S;
        S.verbosity = 0; // Used to be set to the variable verb. Patrik set verbosity manually here.
        
        solver = &S;
        // Use signal handlers that forcibly quit until the solver will be able to respond to
        // interrupts:
        signal(SIGINT, SIGINT_exit);
        signal(SIGXCPU,SIGINT_exit);

        // Set limit on CPU-time:
        if (cpu_lim != INT32_MAX){
            rlimit rl;
            getrlimit(RLIMIT_CPU, &rl);
            if (rl.rlim_max == RLIM_INFINITY || (rlim_t)cpu_lim < rl.rlim_max){
                rl.rlim_cur = cpu_lim;
                if (setrlimit(RLIMIT_CPU, &rl) == -1)
                    printf("WARNING! Could not set resource limit: CPU-time.\n");
            } }

        // Set limit on virtual memory:
        if (mem_lim != INT32_MAX){
            rlim_t new_mem_lim = (rlim_t)mem_lim * 1024*1024;
            rlimit rl;
            getrlimit(RLIMIT_AS, &rl);
            if (rl.rlim_max == RLIM_INFINITY || new_mem_lim < rl.rlim_max){
                rl.rlim_cur = new_mem_lim;
                if (setrlimit(RLIMIT_AS, &rl) == -1)
                    printf("WARNING! Could not set resource limit: Virtual memory.\n");
            } }
 
 
        // Initially set the reordering type to 'none', i.e. no reordering. This can be changed with a 
        // REORDER command.        
        setReorderType(0);
                                                                                          
        // Starting command-response loop. Basically, we keep reading commands from stdin forever,
        // until we see a "QUIT" command.
        int readyToQuit = 0;      
        char cmd[MAXCMDLEN];
        while (!readyToQuit)
        {
          // Read the next command
          getNextCmd(cmd);
 
          // Print out the command
          int cmdnum;
          char cmdtype[MAXCMDLEN], reordertype[MAXCMDLEN];                                    
          sscanf(cmd, "%d %s", &cmdnum, cmdtype);

          // Quit
          if (strcmp(cmdtype,"QUIT")==0)
          {
            char response[100] = "OK";
            writeResponse(cmdnum, response);
            exit(0);
          }
          // Reorder
          else if (strcmp(cmdtype,"REORDER")==0)
          {           
            sscanf(cmd, "%d %s %s", &cmdnum, cmdtype, reordertype);    
            if (cmdnum != 1)
            {
              fprintf(stderr,"REORDER has to be first command. Exiting.\n");
              exit(1);              
            }
                                  
            if (strcmp(reordertype,"none")==0) setReorderType(0);
            else if (strcmp(reordertype,"array")==0) setReorderType(1);
            else if (strcmp(reordertype,"tree")==0) setReorderType(2);
            else 
            {
              fprintf(stderr,"Unknown REORDER type! Exiting.\n");
              exit(1);
            }
            char response[100] = "OK";
            writeResponse(cmdnum, response);
          }
          // Load a constraint file
          else if (strcmp(cmdtype,"LOAD")==0)
          {
            // Actually do the load into minisat  
            parse_DIMACS(S);
            char response[100] = "OK";
            writeResponse(cmdnum, response);
          }
          // Single solution or backbone computation
          else if ((strcmp(cmdtype,"SOLVE")==0) || (strcmp(cmdtype,"BACKBONE")==0))
          {
            // Change to signal-handlers that will only notify the solver and allow it to terminate voluntarily:
            signal(SIGINT, SIGINT_interrupt);
            signal(SIGXCPU,SIGINT_interrupt);
            
            // Actually perform the solving!
            vec<Lit> dummy;
            lbool ret = S.solveLimited(dummy);

            // Only if there actually was a solution is there any point in going forward!
            if (ret == l_True)
            {
              // Print that it was satisfiable 
              printf("SAT\n");
              fflush(NULL);

              // Go through all the specified variables
              int addspace = 0;
              char tempbuffer[1000];
              int atendofline = 0;
              while (!atendofline)
              {
                  // Read the variable list up to and including the first comma (',') or dash '-' or end of line
                char c;
                char *ptr = tempbuffer;
                while (1)
                {
                  c = fgetc(stdin);
                  if ((c < '0') || (c > '9')) 
                  {
                    *ptr = '\0';
                    break;
                  }
                  *(ptr++) = c;
                } 

                long long rangeMin, rangeMax; 
                sscanf(tempbuffer, "%lld", &rangeMin );

                if (c == '-')
                {
                    // dash, so we should read the range maximum as well
                  ptr = tempbuffer;
                  while (1)
                  {
                    c = fgetc(stdin);
                    if ((c < '0') || (c > '9')) 
                    {
                      *ptr = '\0';
                      break;
                    }
                    *(ptr++) = c;
                  } 
                  sscanf(tempbuffer, "%lld", &rangeMax );
                }
                else
                {
                    // not a dash, so just a single value
                  rangeMax = rangeMin;
                }

                if (c == '\n')
                {
                    // came to the end of the line, so mark this
                  atendofline = 1; 
                }

                for (long long externVar = rangeMin; externVar <= rangeMax; externVar++)
                {
                  int internVar = getInternVar(externVar);

                  if (strcmp(cmdtype,"SOLVE")==0)
                  {
                    if (internVar)
                    {
                      if (addspace) printf(" "); else addspace=1;                     
                      if (S.model[internVar-1] != l_Undef)
                      {
                        printf("%s%lld", (S.model[internVar-1]==l_True)?"+":"-", externVar);
                      }
                      else
                      {
                        printf("?%lld", externVar);                       
                      }
                    }
                  }
                  
                  if (strcmp(cmdtype,"BACKBONE")==0)
                  {           
                    if (internVar)
                    {
                      if (addspace) printf(" "); else addspace=1;
                      if (S.model[internVar-1] != l_Undef)
                      {
                        // Is the value opposite to that given in the found solution possible?
                        // NOTE: In the code below it would at first seem that I am testing true-true and false-false, rather than
                        // true-false and false-true. But this is the way to get it to work.
                        if (S.model[internVar-1]==l_True)
                        {
                          if (S.solve(mkLit(internVar-1, true)))
                          {
                            // Solution found, not backbone.
                            printf("%lld", externVar);
                          }
                          else
                          {
                            // Solution NOT found, so this is a backbone (true) variable
                            printf("%s%lld", "+", externVar);                            
                          } 
                        }
                        else if (S.model[internVar-1]==l_False)
                        {
                          if (S.solve(mkLit(internVar-1, false)))
                          {
                            // Solution found, not backbone.
                            printf("%lld", externVar);
                          }
                          else
                          {
                            // Solution NOT found, so this is a backbone (false) variable
                            printf("%s%lld", "-", externVar);
                          }
                        }
                      }
                      else
                      {
                        printf("?%lld", externVar); 
                      }
                    }
                  }
                }
              }
              printf(" 0\n");
              fflush(NULL);
            }

            else if (ret == l_False)
            {
              printf("UNSAT\n");
              fflush(NULL);
              
              // skip over the next line in the command file (the one that contains the variable ranges)
              char c;
              while ((c = fgetc(stdin)) != '\n') {}
            }
            else
            {
              printf("INDET\n");
              fflush(NULL);

              // skip over the next line in the command file (the one that contains the variable ranges)
              char c;
              while ((c = fgetc(stdin)) != '\n') {}
            }

            char response[100] = "OK";
            writeResponse(cmdnum, response);

          }
        }
                
    } catch (OutOfMemoryException&){
        printf("INDETERMINATE\n");
        exit(0);
    }
}
